# -*- coding: utf-8 -*-
"""
Created on Sun May 26 11:21:31 2019

@author: yoelr
"""
from biosteam.evaluation import Model, Metric
from biosteam.evaluation.evaluation_tools import triang
import biosteam.biorefineries.sugarcane as sc

__all__ = ('sugarcane_model',)

tea = sc.sugarcane_sys.TEA
ethanol = sc.system.ethanol
get_prodcost = lambda: float(tea.production_cost(ethanol))
get_FCI = lambda: tea._FCI_cached
get_prod = lambda: ethanol.massnet * tea._annual_factor

BT = sc.system.BT
sc_sys = sc.sugarcane_sys
def get_steam():
    return sum([i.flow for i in BT.steam_utilities])*18.01528*tea._annual_factor/1000

power_utils = ([i._power_utility for i in sc_sys.units if (i._power_utility and i is not BT)])
excess_electricity = [0]
def get_consumed_electricity():
    factor = tea._annual_factor/1000
    electricity_generated = -BT._power_utility.rate * factor
    consumed_electricity = sum([i.rate for i in power_utils]) * factor
    excess_electricity[0] = electricity_generated - consumed_electricity
    return consumed_electricity
get_excess_electricity = lambda: excess_electricity[0]

metrics = (Metric('Internal rate of return', sc.sugarcane_tea.solve_IRR, '%'),
           Metric('Ethanol production cost', get_prodcost, 'USD/yr'),
           Metric('Fixed capital investment', get_FCI, 'USD'),
           Metric('Ethanol production', get_prod, 'kg/hr'),
           Metric('Steam', get_steam, 'MT/yr'),
           Metric('Consumed electricity', get_consumed_electricity, 'MWhr/yr'),
           Metric('Excess electricity', get_excess_electricity, 'MWhr/yr'))

sugarcane_model = Model(sc.sugarcane_sys, metrics, skip=False)
sugarcane_model.load_default_parameters(sc.system.Sugar_cane)
param = sugarcane_model.parameter

# Fermentation efficiency
fermentation = sc.system.P24
@param(element=fermentation, distribution=triang(fermentation.efficiency),
       kind='coupled')
def set_fermentation_efficiency(efficiency):
    fermentation.efficiency= efficiency
    
# Boiler efficiency
BT = sc.system.BT
@param(element=BT, distribution=triang(BT.boiler_efficiency))
def set_boiler_efficiency(boiler_efficiency):
    BT.boiler_efficiency = boiler_efficiency

# Turbogenerator efficiency
@param(element=BT, distribution=triang(BT.turbogenerator_efficiency))
def set_turbogenerator_efficiency(turbo_generator_efficiency):
    BT.turbo_generator_efficiency = turbo_generator_efficiency
    
# RVF separation
rvf = sc.system.P14
@param(element=rvf, distribution=triang(rvf.split['Lignin']),
        kind='coupled')
def set_rvf_solids_retention(solids_retention):
    rvf.split['Lignin', 'CaO', 'Ash', 'Cellulose', 'Hemicellulose'] = solids_retention












#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Mar 10 16:26:31 2018

All species for the oil and sugar separation (pretreatment) section of the lipid cane baseline biorefinery are defined here in the Species object, pretreatment_species.

@author: Yoel Rene Cortes-Pena
"""
from biosteam import Species
from biosteam.compounds import Substance
from .ethanol import ethanol_species
from .biodiesel import lipid

__all__ = ['pretreatment_species']

sp = Species()
for name in ('Ash', 'Cellulose', 'Flocculant', 'Hemicellulose', 'Lignin', 'Solids'):
    setattr(sp, name, Substance(name, rho=1540))
    
sp.Lipid = lipid
sp.Lipid.ID = 'Lipid'
sp.CaO = Substance('CaO', MW=56.0774, rho=1540)
sp.setprops(sp.IDs, 'Hf', 0)
sp.setprops(sp.IDs, 'T_ref', 298.15)

# %% Heat capacities

# References
# https://www.sciencedirect.com/science/article/pii/0032386182901252
# https://link.springer.com/article/10.1007%2Fs10853-013-7815-6

pretreatment_species = Species.tospecies(tuple(ethanol_species)+tuple(sp))
sp = pretreatment_species
sp.Ash.Cp = sp.Ash.Cpm = 0.09*4184/1000  # from Kumar
sp.CaO.Cp = 1023.88/1000  # from Kumar
sp.CaO.Cpm = 1023.88/1000 * 56.0774
sp.Cellulose.Cp = sp.Cellulose.Cpm = 1364/1000  # from paper
sp.Hemicellulose.Cp = sp.Hemicellulose.Cpm = 1364/1000  # same as cellulose
sp.Flocculant.Cp = sp.Flocculant.Cpm = 4184/1000  # from Kumar
sp.Lignin.Cp = sp.Lignin.Cpm = 1364/1000  # same as cellulose
sp.Solids.Cp = sp.Solids.Cpm = 1100/1000  # common value for solids

sp.setprops(['Ash', 'CaO', 'Cellulose', 'Flocculant',
             'Hemicellulose', 'Lignin', 'Solids'], 'Hvapm', 0)
sp.IDs =  ('Glucose', 'H3PO4', 'Flocculant',
           'Ethanol', 'Lignin', 'Solids', 'Sucrose',
           'CaO', 'Ash', 'Cellulose', 'Hemicellulose',
           'Lipid', 'Water')

# %% Heat of combustion

# kJ/kmol
sp.Hemicellulose.Hc = 17000
sp.Cellulose.Hc = 17000
sp.Lignin.Hc = 21000
sp.Glucose.Hc = 2800e3
sp.Sucrose.Hc = 5700e3
sp.Lipid.Hc = 36e6


# The dry heat values of cellulose, hemicellulose and lignin respectively are 17 MJ/kg (7,320 Btu/lb), 16.63 MJ/kg (7165 Btu/lb) and 21.13 MJ/kg (9,105 Btu/lb) [1]

# [1] Murphy W. K., and K. R. Masters. 1978. Gross heat of combustion of northern red oak (Quercus rubra) chemical components. Wood Sci. 10:139-141.






# -*- coding: utf-8 -*-
"""
Created on Tue Aug 13 21:43:56 2019

@author: yoelr
"""
from biosteam.evaluation import evaluation_tools as tools
from biosteam.evaluation import Model, Metric
from biosteam.biorefineries.cornstover.system import \
    cornstover_sys, ethanol_tea, cornstover_tea, \
    ethanol, cornstover, R301, ethanol_density_kggal, \
    areas, BT, Area700

get_MESP = lambda: cornstover_tea.solve_price(ethanol, ethanol_tea) * ethanol_density_kggal
get_FCI = lambda: sum([i._FCI_cached for i in cornstover_tea.TEAs])
get_coproduct_credit = lambda: sum([i._utility_cost_cached for i in cornstover_tea.TEAs])
get_ethanol_sales = lambda: ethanol.cost * ethanol_tea._annual_factor
pws = [i._power_utility for i in cornstover_sys.units
       if i._power_utility and i._power_utility.rate > 0]
get_excess_electricity = lambda: BT._Design['Work'] - sum([i.rate for i in pws])

metrics =[Metric('Minimum ethanol selling price', get_MESP, 'USD/gal'),
          Metric('Fixed capital investment', get_FCI, 'USD'),
          Metric('Co-product credit', get_coproduct_credit, 'USD/yr'),
          Metric('Ethanol sales', get_ethanol_sales, 'USD/yr'),
          Metric('Excess electricity', get_excess_electricity, 'kW')]

def electricity_rate_function(tea):
    power_utilities = [i._power_utility for i in tea.units if i._has_power_utility]
    if tea is Area700:
        boiler_item = BT.cost_items['Boiler']
        Design = BT._Design
        return lambda: boiler_item.kW/boiler_item.S * Design['Flow rate']/1e3
    return lambda: sum([i.rate for i in power_utilities])/1e3

def cooling_duty_function(tea):
    heat_utilities = sum([i._heat_utilities for i in tea.units if i._N_heat_utilities], [])
    cooling_utilities = [i for i in heat_utilities if i.duty < 0]
    return lambda: sum([i.duty for i in cooling_utilities])

def installation_cost_function(tea):
    return lambda: tea.installation_cost

for i, tea in enumerate(areas, 1):
    Area = f'Area {i}00'
    metrics.extend(
        (Metric('Electricity', electricity_rate_function(tea), 'MW', Area),
         Metric('Cooling duty', cooling_duty_function(tea), 'MMkcal/hr', Area),
         Metric('Installation cost', installation_cost_function(tea), '10^6 USD', Area)))

cornstover_model = Model(cornstover_sys, metrics)
cornstover_model.load_default_parameters(cornstover)
param = cornstover_model.parameter

# Add saccharification as a parameter
saccharification_reaction = R301.saccharification[2]
X = tools.bounded_triang(saccharification_reaction.X, addition=0.05)
@param(element=R301, kind='coupled', distribution=X)
def set_saccharification_conversion(saccharification_conversion):
    saccharification_reaction.X = saccharification_conversion

# Add ethanol conversion as a parameter
ethanol_reaction = R301.cofermentation[0]
X = tools.bounded_triang(ethanol_reaction.X, addition=0.05)
@param(element=R301, kind='coupled', distribution=X)
def set_ethanol_conversion(ethanol_conversion):
    ethanol_reaction.X = ethanol_conversion
    
# Add saccharification time as a parameter
X = tools.triang(R301.tau_saccharification)
@param(element=R301, kind='isolated', distribution=X)
def set_saccharification_time(saccharification_time):
    R301.tau_saccharification= saccharification_time
    
# Add fermentation time as a parameter
X = tools.triang(R301.tau_cofermentation)
@param(element=R301, kind='isolated',  distribution=X)
def set_fermentation_time(fermentation_time):
    R301.tau_cofermentation = fermentation_time




# -*- coding: utf-8 -*-
"""
Created on Wed Mar  6 16:46:04 2019

@author: yoelr
"""

__all__ = []

from . import _vacuum
from . import _tables
from . import _cost_index
from . import _batch

from ._vacuum import *
from ._tables import *
from ._cost_index import *
from ._batch import *

__all__.extend(_cost_index.__all__)
__all__.extend(_vacuum.__all__)
__all__.extend(_tables.__all__)
__all__.extend(_batch.__all__)


# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 13:50:03 2018

@author: yoelr
"""
from thermo.chemical import Chemical as TChem
from ._compound import Compound
import numpy as np
import re

__all__ = ('Chemical', 'Solid', 'Liquid', 'Gas')

class Chemical(Compound, TChem):
    """An extension of the ChEDL thermo Chemical class. The enthalpy property, 'H', does not account for excess ethalpy, it is simply based on latent heats and heat capacities at a constant pressure of 101325. All thermodynamic energies are now calculated properties and not attributes. 

`Read the docs for thermo.Chemical for accurate documentation. <http://thermo.readthedocs.io/en/latest/thermo.chemical.html>`__"""

    def __init__(self, ID, T=298.15, P=101325):
        try:
            TChem.__init__(self,
                           re.sub(r"\B([A-Z])", r" \1", ID).capitalize().replace('_', ' '),
                           T, P)
        except:
            raise LookupError(f"chemical '{ID}' not found in data bank, try chemical's CAS")
        self.ID = ID.replace(' ', '_')
        if self.CAS == '56-81-5':
            self.__UNIFAC_Dortmund_groups = {2: 2, 3: 1, 14: 2, 81: 1}
        self.Hfm = self.Hf

    # These properties are gone
    Hm = None
    Sm = None
    Um = None
    Am = None
    
    # New units for heat capacity
    @property
    def Cp(self):
        "Specific heat capacity (kJ/kg/K)"
        return self.Cpm/self.MW

    # Baseline equation of state set at atmospheric pressure
    def set_thermo(self):
        try: self._eos_T_101325 = self.eos.to_TP(self.T, 101325)
        except: pass


class Solid(Chemical):
    """Create a :doc:`Chemical` such that its phase remains as solid."""
    def __init__(self, ID, T=298.15, P=101325):
        super().__init__(ID, T, P)
        # Phase change temperature consitent with phase
        self.Tb = np.inf 
        self.Tm = np.inf
        self.phase_ref = 's'
    
    @property
    def phase(self): return 's'
    @phase.setter
    def phase(self, phase): pass


class Liquid(Chemical):
    """Create a :doc:`Chemical` such that its phase remains as liquid."""
    phase_ref = 'l'
    def __init__(self, ID, T=298.15, P=101325):
        super().__init__(ID, T, P)
        # Phase change temperature consitent with phase
        self.Tb = np.inf 
        self.Tm = 0
        self.phase_ref = 'l'
    
    @property
    def phase(self): return 'l'
    @phase.setter
    def phase(self, phase): pass


class Gas(Chemical):
    """Create a :doc:`Chemical` such that its phase remains as gas."""
    phase_ref = 'g'
    def __init__(self, ID, T=298.15, P=101325):
        super().__init__(ID, T, P)
        # Phase change temperature consitent with phase
        self.Tb = 0
        self.Tm = 0
        self.phase_ref = 'g'
    
    @property
    def phase(self): return 'g'
    @phase.setter
    def phase(self, phase): pass






# -*- coding: utf-8 -*-
"""
Created on Thu Jun 27 23:12:04 2019

@author: yoelr
"""
import biosteam as bst
import biosteam.compounds as cp
from biosteam.biorefineries.lipidcane import species as lcspecies
import pandas as pd
from copy import copy as _copy

__all__ = ('species', 'get_grouped_species', 'species_groups')

# %%  Constants

# Common structural carbohydrates properties

# Assume heat capacity of lignin, cellulose, and hemicellulose
# and all components at 350 K are about the same [2,2].
Cp_cellulosic = 1.364

# Heat of combustions of lignocellulosic material are
# based on wood as an approximiation [4]. Also, assume
# structural carbohydrates have the same heat of combustion
# as cellulose.
# These properties match NREL's
Hc_lignin = 21e3 # (J/g)
Hc_cellulosic = 17000 # (J/g)
cal2joule = 4.184

# %% Initialize species object and define functions

lcspecies = bst.Species.tospecies([*lcspecies.biodiesel_species,
                                   *lcspecies.ethanol_species,
                                   *lcspecies.pretreatment_species])

sp = bst.Species()

def remove(ID):
    try: missing.remove(ID)
    except ValueError: raise ValueError(f"'{ID}' not a required species")

def addspecies(*IDs, cls=None, **kwargs):
    if cls is None: cls = bst.Chemical
    sp.extend([cls(i, **kwargs) for i in IDs])
    for ID in IDs: remove(ID)

def addchemical(ID, ChemID=None, phase=None, **properties): 
    setattr(sp, ID, cp.StaticChemical(ChemID or ID, phase=phase, **properties))
    remove(ID)
    
def addsubstance(ID, *args, **kwargs):
    setattr(sp, ID, cp.Substance(ID, *args, **kwargs))
    remove(ID)

def addfrom(species, ID, species_ID=None):
    if not species_ID: species_ID = ID
    cmp = _copy(getattr(species, species_ID))
    setattr(sp, ID, cmp)
    cmp.ID = species_ID
    remove(ID)

def copy(ID, ChemID):
    cmp = _copy(getattr(sp, ChemID))
    setattr(sp, ID, cmp)
    cmp.ID = ChemID
    remove(ID)

# Species to define
species_IDs = [
        'H2O', 'Ethanol', 'Glucose', 'Galactose',
        'Mannose', 'Xylose', 'Arabinose', 'Cellobiose',
        'Sucrose', 'GlucoseOligomer', 'GalactoseOligomer',
        'MannoseOligomer', 'XyloseOligomer', 'ArabinoseOligomer',
        'Extract','SolubleLignin','HMF', 'Furfural', 'AceticAcid',
        'LacticAcid', 'Xylitol', 'Glycerol', 'SuccinicAcid',
        'NH3', 'H2SO4', 'NH4SO4', 'AmmoniumAcetate', 'DAP',
        'HNO3', 'NaNO3', 'NaOH', 'CellulaseNutrients',
        'Denaturant', 'Oil', 'Cellulose', 'Galactan', 'Mannan', 'Glucan',
        'Xylan', 'Arabinan', 'Lignin', 'Acetate', 'Protein',
        'Ash', 'Enzyme', 'DenaturedEnzyme', 'Z_mobilis', 'T_reesei',
        'Biomass', 'Tar', 'CaO', 'CaSO4', 'Graphite', 'N2', 'O2', 'CO2',
        'CH4', 'H2S', 'SO2', 'NO', 'CO', 'AmmoniumSulfate', 'NO2', 'CSL',
        'WWTsludge', 'Cellulase'
]
missing = species_IDs.copy()

# %% Define species

# TODO: Add heats of combustion

# As is in data bank
addspecies('H2O', 'Ethanol', 'AceticAcid', 'Furfural', 'Glycerol',
           'H2SO4', 'LacticAcid', 'SuccinicAcid') 
sp.SuccinicAcid.Hf = -940.26e3 # kJ/mol
sp.LacticAcid.T = sp.LacticAcid.Tb = 122+273.15
sp.LacticAcid.H_int_l_T_ref_l_to_Tb = sp.LacticAcid.H
sp.LacticAcid.Hvap_Tbm = sp.LacticAcid.Hvap
sp.LacticAcid.Hf = -163122*cal2joule
addchemical('HNO3', 'NitricAcid')
addchemical('Denaturant', 'Octane')
addchemical('DAP', 'Diammonium Phosphate')
addchemical('AmmoniumAcetate')
addchemical('NH4SO4', 'AmmoniumSulfate')
addchemical('NaNO3', 'SodiumNitrate')
addchemical('HMF')

# Will remain in the vapor phase
addspecies('N2', 'NH3', 'O2', 'CH4', 'H2S', 'SO2', cls=cp.StaticChemical)
addchemical('CO2', phase='g')

# Analagous vapors
addsubstance('NO2', MW=46.01, obj=sp.N2, Hf=7925*cal2joule)
addsubstance('NO', MW=30.01, obj=sp.N2, Hf=82.05)
addsubstance('CO', MW=28.01, obj=sp.N2, Hf=-110.522)

# Will remain as  solid
addspecies('Glucose', 'Xylose', 'Sucrose',
           cls=cp.StaticChemical, Cp=Cp_cellulosic)
sp.Glucose.Hf = -300428*cal2joule
sp.Xylose.Hf = -249440*cal2joule
sp.Sucrose.Hf = -480900*cal2joule
addspecies('CaSO4', 'Graphite', 'AmmoniumSulfate',
           cls=cp.StaticChemical, Cp=Cp_cellulosic)

# Analagous sugars
copy('Mannose', 'Glucose')
copy('Galactose', 'Glucose')
copy('Arabinose', 'Xylose')

# Other analogues
copy('CellulaseNutrients', 'Glucose')
copy('Extract', 'Glucose')
copy('Acetate', 'AceticAcid')
sp.Acetate.Hf = -103373
copy('Tar', 'Xylose')

# Species taken from previous study
addchemical('Oil', 'Oleic Acid', phase='l')
addfrom(lcspecies, 'CaO')
addfrom(lcspecies, 'Ash')
addfrom(lcspecies, 'NaOH')
addsubstance('Lignin', Cp=Cp_cellulosic, Hf=-108248*cal2joule, MW=152.15)
copy('SolubleLignin', 'Lignin')

# Create structural carbohydrates
addsubstance('GlucoseOligomer', obj=sp.Glucose, Cp=Cp_cellulosic,
             MW=162.1424, Hf=-233200*cal2joule)
copy('GalactoseOligomer', 'GlucoseOligomer')
copy('MannoseOligomer', 'GlucoseOligomer')
addsubstance('XyloseOligomer', obj=sp.Xylose, Cp=Cp_cellulosic,
             MW=132.11612, Hf=-182100*cal2joule)
copy('ArabinoseOligomer', 'XyloseOligomer')

# Other
addsubstance('Z_mobilis', MW=24.6265, Hf=-31169.39*cal2joule)
addsubstance('T_reesei', MW=23.8204, Hf=-23200.01*cal2joule)
addsubstance('Biomass', MW=23.238, Hf=-23200.01*cal2joule)
addsubstance('Cellulose', MW=162.1406, Hf=-233200.06*cal2joule, Hc=Hc_cellulosic)
addsubstance('Protein', MW=22.8396, Hf=-17618*cal2joule)
addsubstance('Enzyme', MW=24.0156, Hf=-17618*cal2joule)
addsubstance('Glucan', MW=162.14, Hf=-233200*cal2joule)
addsubstance('Xylan', MW=132.12, Hf=-182100*cal2joule)
addsubstance('Xylitol', MW=152.15, Hf=-243145*cal2joule)
addsubstance('Cellobiose', MW=342.30, Hf=-480900*cal2joule)
addsubstance('CSL', MW=1, Hf=sp.Protein.Hf/4+sp.H2O.Hf/2+sp.LacticAcid.Hf/4)
copy('DenaturedEnzyme', 'Enzyme')
copy('Arabinan', 'Xylan')
copy('Mannan',   'Glucan')
copy('Galactan', 'Glucan')

# %% TODO: Maybe remove this

# missing.extend([
#     'OtherSugars', 'SugarOligomers', 
#     'InorganicSolubleSolids',
#     'Furfurals', 'OtherOrganics',
#     'OtherOrganics', 'OtherOrganics',
#     'COxSOxNOxH2S', 'OtherStructuralCarbs',
#     'CellMass',
#     'OtherInsolubleSolids',
#     'OrganicSolubleSolids',
# ])
# copy('OtherSugars', 'Arabinose')
# copy('SugarOligomers', 'GlucoseOligomer')
# copy('OrganicSolubleSolids', 'SolubleLignin')
# copy('InorganicSolubleSolids', 'DAP')
# copy('Furfurals', 'HMF')
# copy('OtherOrganics', 'Denaturant')
# copy('COxSOxNOxH2S', 'NO')
# copy('OtherStructuralCarbs', 'Arabinan')

# For waste water
copy('WWTsludge', 'Z_mobilis')
copy('Cellulase', 'Enzyme')
# copy('OtherInsolubleSolids', 'Tar')
bst.Stream.species = species = sp

# %% Grouped species

species_groups = dict(
    OtherSugars = ['Arabinose',
                   'Mannose',
                   'Galactose',
                   'Cellobiose',
                   'Sucrose'],
    SugarOligomers = ['GlucoseOligomer',
                      'XyloseOligomer',
                      'GalactoseOligomer',
                      'ArabinoseOligomer',
                      'MannoseOligomer'],
    OrganicSolubleSolids = ['AmmoniumAcetate',
                            'SolubleLignin',
                            'Extract', 
                            'LacticAcid', 
                            'Cellulase'],
    InorganicSolubleSolids = ['AmmoniumSulfate',
                            'DAP',
                            'NaOH',
                            'HNO3',
                            'NaNO3'],
    Furfurals = ['Furfural',
                 'HMF'],
    OtherOrganics = ['Glycerol',
                     'Denaturant',
                     'Oil',
                     'SuccinicAcid',
                     'Xylitol'],
    COxSOxNOxH2S = ['NO',
                    'NO2',
                    'SO2',
                    'CO',
                    'H2S'],
    Protein = ['Protein',
               'Enzyme',
               'DenaturedEnzyme'],
    CellMass = ['WWTsludge',
                'Z_mobilis',
                'T_reesei'],
    OtherInsolubleSolids = ['Tar',
                            'Ash',
                            'Graphite',
                            'Lime'],
    OtherStructuralCarbohydrates = ['Arabinan', 
                                    'Mannan', 
                                    'Galactan']
)

def get_grouped_species(stream, units='kmol/hr'):
    s = bst.Stream(species=species)
    s.setflow(flow=stream.mol, species=stream.species.IDs)
    return pd.Series({i:s.getflow(*j, units=units).sum() for i, j in species_groups.items()})

assert not missing, str(missing)

# Fix sugar properties


# cellulosic = ('Glucan', 'Xylan', 'GlucoseOligomer','XyloseOligomer',
#               'Acetate', 'SolubleLignin')




# -*- coding: utf-8 -*-
"""
Created on Wed May 22 23:14:02 2019

@author: yoelr
"""
import biosteam as bst
from .._exceptions import DimensionError

__all__ = ('DisplayUnits',)

setattr_ = object.__setattr__
getattr_ = object.__getattribute__

class DisplayUnits:
    """Create a DisplayUnits object where default units for representation are stored."""
    def __init__(self, **display_units):
        setattr_(self, '_display_units', display_units)
        list_keys = []
        for k, v in display_units.items():
            try: # Assume units is one string
                dims = getattr(bst._ureg, v).dimensionality
            except:
                try: # Assume units are a list of possible units
                    dims = [getattr(bst._ureg, i).dimensionality for i in v]
                    list_keys.append(k)
                except: # Assume the user uses value as an option, and ignores units
                    dims = v
            setattr_(self, k, dims)
        for k in list_keys:
            display_units[k] = display_units[k][0] # Default units is first in list
    
    def __setattr__(self, name, unit):
        if name not in self.__dict__:
            raise AttributeError(f"can't set display units for '{name}'")
        if not isinstance(unit, str) and isinstance(unit, type(self._display_units[name])):
            self._display_units[name] = unit
            return
        name_dim = getattr_(self, name)
        unit_dim = getattr(bst._ureg, unit).dimensionality
        if isinstance(name_dim, list):
            if unit_dim not in name_dim:
                raise DimensionError(f"dimensions for '{name}' must be either {', '.join(name_dim[:-1])} or {name_dim[-1]}; not {unit_dim}")    
        else:
            if name_dim != unit_dim:
                raise DimensionError(f"dimensions for '{name}' must be in {name_dim}, not {unit_dim}")
        self._display_units[name] = unit
    
    def __getattribute__(self, name):
        if name[0] == '_':
            return getattr_(self, name)
        else: 
            return self._display_units[name]
    
    def __iter__(self):
        for i in self._display_units.values():
            if isinstance(i, list): yield i[0]
            yield i
            
    def __repr__(self):
        sig = ', '.join((f"{i}='{j}'" if isinstance(j, str) else f'{i}={j}') for i,j in self._display_units.items())
        return f'{type(self).__name__}({sig})'
        


# -*- coding: utf-8 -*-
"""
The biosteam.reaction packages features objects to manage stoichiometric reactions and conversions. For a complete example, check out :doc:`stoichiometric reactions <../tutorial/Stoichiometric reactions>`.
"""

__all__ = []

from . import _reaction

from ._reaction import *

__all__.extend(_reaction.__all__)


# -*- coding: utf-8 -*-
"""
Created on Tue Mar 19 13:20:52 2019

Module for estimating utility price based on:

How to Estimate Utility Costs. Engineering Practice. CHEMICAL ENGINEERING. WWW.CHE.COM, APRIL 2006. Gael D. Ulrich and Palligarnai T. VasudevanUniversity of New Hampshire

https://terpconnect.umd.edu/~nsw/chbe446/HowToEstimateUtilityCosts-UlrichVasudevan2006.pdf
    
Utility estimates are often complicated because they depend on both inflation and energy costs. This simplified approach offers a two-factor utility-cost equation and the relevant coefficients for a number of utilities. 

@author: yoelr
"""
import numpy as np

ln = np.log


# %% Classes to calculate utilities

__all__ = ('Utility', 'Electricity', 'CoolingWater')

class Utility:
    """Abstract class for utilities."""
    __slots__ = _attr2show_ = ()
    ID = 'Utility'
    fuelprice = 12 #: $/GJ
    CEPCI = 567.5 #: Chemical Engineering Plant Cost Index
    units = {}
        
    @property
    def price(self):
        """Price of utility."""
        return self._a * self.CEPCI + self._b * self.fuelprice
    
    def __repr__(self):
        units = self.units.get('price')
        if units: units = ' ' + units 
        return f"<{self.ID}: price={self.price:.2g}{units}>"
    
    def _info(self):
        units = self.units.get('price')
        if units: units = ' ' + units
        info =(f"{self.ID}:\n"
             + f" price: {self.price:.2g}{units}")  
        for attr in self._attr2show_:
            value = getattr(self, attr)
            units = self.units.get(attr, '')
            if units:
                units = ' ' + units
                info += f"\n {attr}: {value:.2g}{units}"
            else:
                info += f"\n {attr}: {value}{units}"
        return info
    
    def show(self):
        print(self._info())
    

class Electricity(Utility):
    __slots__ = _attr2show_ = ('purchased', 'grassroots')
    ID = 'Electricity'
    units = {'price': 'USD/kWhr'}
    
    def __init__(self, purchased=True, grassroots=False):
        self.purchased = purchased
        self.grassroots = grassroots
    
    @property
    def _a(self):
        if self.purchased:
            return 1.3e-4
        elif self.grassroots:
            return 1.1e-4
        else:
            return 1.3e-4
    
    @property
    def _b(self):
        if self.purchased:
            return 0.01
        else:
            return 0.011

class CoolingWater(Utility):
    
    __slots__ = ('grassroots', '_flow', '_b')
    _attr2show_ = ('flow', 'grassroots')
    ID = 'Cooling water'
    units = {'price': 'USD/m^3',
             'flow': 'm^3/s'}
    _flowbounds = (0.01, 10)
    
    def __init__(self, flow, grassroots=True):
        self.grassroots = grassroots
        self.flow = flow
        self._b = 0.003
    
    @property
    def _a(self):
        if self.grassroots:
            return 7e-5 + 2.5e-5 * self._flow**-1 
        else:
            return 1e-4 + 3.0e-5 * self._flow**-1
    
    @property
    def flow(self):
        """Volumetric flow rate (m3/s)."""
        return self._flow
    
    @flow.setter
    def flow(self, flow):
        lb, ub = self._flowbounds
        if not lb <= flow <= ub:
            raise ValueError("flow must be between 0.01 and 10 m3/s")
        self._flow = flow
        
    
    
    
    
    
    
    
    
    
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Sat Sep  1 17:35:28 2018

@author: yoelr
"""
import numpy as np
from .. import Stream, Unit
from ..utils import MissingStream

__all__ = ('MassBalance',)

# %% Mass Balance Unit

class MassBalance(Unit):
    """Create a Unit object that changes net input flow rates to satisfy output flow rates. This calculation is based on mass balance equations for specified species. 

    Parameters
    ----------
    species : tuple[str]
        Species that will be used to solve mass balance linear equations. The number of species must be same as the number of input streams varied.
    streams : tuple[int]
        Indices of input streams that can vary in net flow rate
    exact=True : bool, optional
        True if exact flow rate solution is required for the specified species.
    balance='flow' : {'flow', 'fraction'}, optional
          * 'flow': Satisfy output flow rates
          * 'fraction': Satisfy net output molar fractions

    Examples
    --------
    :doc:`notebooks/MassBalance Example`
    
    """
    line = 'Balance'
    results = None
    _has_cost = False
    _N_outs = 0

    def __init__(self, ID='', outs=(), ins=None,
                 species=None, streams=None,
                 exact=True, balance='flow'):
        self.ID = ID
        self._init_ins(ins)
        self._init_outs(outs)
        self._balance = balance
        self._streams = streams
        self._species = species
        
        # Make sure balance type is valid
        if balance not in ('flow', 'fraction'):
            raise ValueError(
                "balance type must be one of the following: 'flow', 'fraction'")

        # Catch correct solver and make sure linear system of equations is square to achieve exact solution
        if exact:
            solver = np.linalg.solve
            species_IDs = species
            sID, s_index = len(species_IDs), len(streams)
            if sID != s_index:
                raise ValueError(
                    f"length of species ({sID}) must be equal to the length of streams_index ({s_index}) when exact solution is needed")
        else:
            solver = np.linalg.lstsq

        # Cach indices for species and solver
        self._bal_index = Stream.indices(species)
        self._linalg_solver = solver

    def _init_ins(self, ins):
        """Initialize input streams."""
        if ins is None:
            self._ins = [MissingStream for i in range(self._N_ins)]
        elif isinstance(ins, Stream):
            self._ins = [ins]
        elif isinstance(ins, str):
            self._ins = [Stream(ins)]
        elif not ins:
            self._ins = [Stream('') for i in range(self._N_ins)]
        else:
            self._ins = [Stream(i) if isinstance(i, str) else i for i in ins]
    
    def _init_outs(self, outs):
        """Initialize output streams."""
        if isinstance(outs, str):
            self._outs = [Stream(outs)]
        elif isinstance(outs, Stream):
            self._outs = [outs]
        elif outs is None:
            self._outs = [MissingStream for i in range(self._N_outs)]
        elif not outs:
            self._outs = [Stream('') for i in range(self._N_outs)]
        else:
            self._outs = [Stream(i) if isinstance(i, str) else i for i in outs]

    def _run(self):
        """Solve mass balance by iteration."""
        # SOLVING BY ITERATION TAKES 15 LOOPS FOR 2 STREAMS
        # SOLVING BY LEAST-SQUARES TAKES 40 LOOPS
        stream_index = self._streams
        balance = self._balance
        solver = self._linalg_solver

        # Set up constant and variable streams
        vary = []  # Streams to vary in mass balance
        constant = []  # Constant streams
        for i in range(len(self.ins)):
            if i in stream_index:
                vary.append(self.ins[i])
            else:
                constant.append(self.ins[i])

        # Species indices used in mass balace
        index = self._bal_index

        if balance == 'flow':
            # Perform the following calculation: Ax = b = f - g
            # Where:
            #    A = flow rate array
            #    x = factors
            #    b = target flow rates
            #    f = output flow rates
            #    g = constant input flow rates

            # Solve linear equations for mass balance
            A = np.array([s.mol for s in vary]).transpose()[index, :]
            f = self._mol_out[index]
            g = sum([s.mol[index] for s in constant])
            b = f - g
            x = solver(A, b)

            # Set flow rates for input streams
            for factor, s in zip(x, vary):
                s.mol[:] = s.mol * factor

        elif balance == 'fraction':
            # Perform the following calculation:
            # Ax = b
            #    = sum( A_ * x_guess + g_ )f - g
            #    = A_ * x_guess * f - O
            # O  = sum(g_)*f - g
            # Where:
            # A_ is flow array for all species
            # g_ is constant flows for all species
            # Same variable definitions as in 'flow'

            # Set all variables
            A_ = np.array([s.mol for s in vary]).transpose()
            A = np.array([s.mol for s in vary]).transpose()[index, :]
            f = self._molfrac_out[index]
            g_ = sum([s.mol for s in constant])
            g = g_[index]
            O = sum(g_) * f - g

            # Solve by iteration
            x_guess = np.ones_like(index)
            not_converged = True
            while not_converged:
                # Solve linear equations for mass balance
                i += 1
                b = (A_ * x_guess).sum()*f + O
                x_new = solver(A, b)
                not_converged = sum(((x_new - x_guess)/x_new)**2) > 0.0001
                x_guess = x_new

            # Set flow rates for input streams
            for factor, s in zip(x_new, vary):
                s.mol = s.mol * factor

    # Setting input and output streams do not change sources or sinks
    @property
    def ins(self):
        """[:] Input streams"""
        return self._ins

    @ins.setter
    def ins(self, streams):
        self._ins[:] = streams

    @property
    def outs(self):
        """[:] Output streams"""
        return self._outs

    @outs.setter
    def outs(self, streams):
        self._outs[:] = streams


# %% Energy Balance Unit

# class EnergyBalance(Unit):
#     """Create a Unit object that changes a stream's temperature, flow rate, or vapor fraction to satisfy energy balance.

#     **Parameters**

#         **index:** [int] Index of stream that can vary in temperature, flow rate, or vapor fraction.
        
#         **Type:** [str] Should be one of the following
#             * 'T': Vary temperature of output stream
#             * 'F': Vary flow rate of input/output stream
#             * 'V': Vary vapor fraction of output stream
        
#         **Qin:** *[float]* Additional energy input.
        
#     .. Note:: This is not a mixer, input streams and output streams should match flow rates.

#     """
#     _kwargs = {'index': None,
#                'Type': 'T',
#                'Qin': 0}
#     line = 'Balance'
#     _has_cost = False
#     _graphics = MassBalance._graphics
#     _init_ins = MassBalance._init_ins
#     _init_outs = MassBalance._init_outs
    
#     def _run(self):        # Get arguments
#         ins = self.ins.copy()
#         outs = self.outs.copy()
#         kwargs = self._kwargs
#         index = kwargs['index']
#         Type = kwargs['Type']
#         Qin = kwargs['Qin']
        
#         # Pop out required streams
#         if Type == 'F':
#             s_in = ins.pop(index)
#             s_out = outs.pop(index)
#         else:
#             s = outs.pop(index)
        
#         # Find required enthalpy
#         H_in = sum(i.H for i in ins) + Qin
#         H_out = sum(o.H for o in outs)
#         H_s = H_out - H_in
        
#         # Set enthalpy
#         if Type == 'T':
#             s.H = -H_s
#         elif Type == 'V':
#             s.enable_phases()
#             s.VLE(Qin=s.H - H_s)
#         elif Type == 'F':
#             s.mol *= (s_out.H - s_in.H)/H_s
#         else:
#             raise ValueError(f"Type must be 'T', 'V' or 'F', not '{Type}'")
            
        
        


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:56:55 2018

@author: yoelr
"""

__all__ = ('size_batch',)

def size_batch(v_0, tau, tau_0, N_reactors, V_wf) -> dict:
    """Solve for batch reactor volume, cycle time, and loading time
    
    Parameters
    ----------
    v_0 
        Flow rate
    tau
        Reaction time
    tau_0
        Turnaround time
    N_reactors
        Number of reactors
    V_wf
        Fraction of working volume
    
    Returns
    -------
    dict[str: float]
        * 'Reactor volume'
        * 'Cycle time'
        * 'Loading time'
        
    """
    V_T = v_0*(tau + tau_0)/(1-1/N_reactors) # Reacting volume
    V_i = V_T/N_reactors
    t_L = V_i/v_0
    t_B = tau + tau_0 + t_L
    V_i /= V_wf
    
    return {'Reactor volume': V_i,
            'Cycle time': t_B, 
            'Loading time': t_L}
        
        


# -*- coding: utf-8 -*-
"""
Created on Sun Sep  1 04:37:16 2019

@author: yoelr
"""
import pandas as pd
import numpy as np
from biosteam.biorefineries import lipidcane as lc
from biosteam.biorefineries import sugarcane as sc

model_lc = lc.model.lipidcane_model
model_sc = sc.model.sugarcane_model
model_lc_lf = lc.model.lipidcane_model_with_lipidfraction_parameter
np.random.seed(1234)
N_spearman_samples = 50
N_coordinate_samples = 10
N_coordinates = 10
rule = 'L'

# %% Monte Carlo across lipid fraction

model_lc.evaluate_across_coordinate(
      lc.utils.set_lipid_fraction, 'Lipid fraction',
      0.11, 0.01, 'Monte Carlo across lipid fraction.xlsx',
      N_coordinate_samples, N_coordinates,  rule, notify=True)
    
# %% Sugar cane Monte Carlo

samples = model_sc.sample(N_coordinate_samples, rule)
model_sc.load_samples(samples)
model_sc.evaluate()
model_sc.table.to_excel('Monte Carlo sugarcane.xlsx')

# %% Spearman's correlation

# samples = model_lc_lf.sample(N_spearman_samples, rule)
# model_lc_lf.load_samples(samples)
# model_lc_lf.evaluate()
# spearman = model_lc_lf.spearman(
#                 metric_names=('Internal rate of return',))
# spearman.to_excel("Spearman correlation lipidcane.xlsx")







# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 21:43:13 2018

@author: yoelr
"""
import numpy as np
import biosteam as bst
from .. import Unit, Stream
from scipy.optimize import brentq
from . import Mixer, HXutility
from ._flash import Evaporator_PV, Evaporator_PQ
from .designtools import vacuum_system
from warnings import warn
import ht
log = np.log
exp = np.exp

# Table 22.32 Product process and design (pg 592)
# Name: ('Area range (m2)', 'Cost(A) (USD)', 'U (kJ/(hr*m2*K)))', 'Material')
evaporators = {'Horizontal tube': 
                    ((9.29, 743.224),
                     lambda A, CE: CE*2.304*A**0.53,
                     4906.02,
                     'Carbon steel'),
               'Long-tube vertical':
                   ((9.29, 743.224),
                    lambda A, CE: CE*3.086*A**0.55,
                    8176.699,
                    'Carbon steel'),
               'Forced circulation': 
                   ((13.935, 8000),
                    lambda A, CE: CE/500*exp(8.2986 + 0.5329*log(A*0.0929)-0.000196*log(A*0.0929)**2),
                    10731.918,
                    'Carbon steel'),
               'Falling film': 
                   ((13.935, 371.612),
                    lambda A, CE: CE*7.416*A**0.55,
                    10220.874,
                    'Stainless steel tubes/Carbon steel shell')}


class MultiEffectEvaporator(Unit):
    """Creates evaporatorators with pressures given by P (a list of pressures). Adjusts first evaporator vapor fraction to satisfy an overall fraction evaporated. All evaporators after the first have zero duty. Condenses the vapor coming out of the last evaporator. Pumps all liquid streams to prevent back flow in later parts. All liquid evaporated is ultimately recondensed. Cost is based on required heat transfer area. Vacuum system is based on air leakage. Air leakage is based on volume, as given by residence time `tau` and flow rate to each evaporator.

    Parameters
    ----------
    component : str
                Component being evaporated.
    P : tuple[float]
        Pressures describing each evaporator (Pa).
    V : float
        Overall molar fraction of component evaporated.
    P_liq : tuple
            Liquid pressure after pumping (Pa).
    
    """
    _units = {'Area': 'm^2',
              'Volume': 'm^3'}
    _has_power_utility = True
    _N_heat_utilities = 2
    BM = 2.45
    line = 'Multi-Effect Evaporator'

    #: Residence time (hr)
    tau = 0.30

    # Evaporator type
    _Type = 'Forced circulation'
    
    # Data for simmulation and costing
    _evap_data = evaporators[_Type]

    @property
    def Type(self):
        """Evaporation type."""
        return self._Type
    @Type.setter
    def Type(self, evap_type):
        try:
            self._evap_data = evaporators[evap_type]
        except KeyError:
            dummy = str(evaporators.keys())[11:-2]
            raise ValueError(f"Type must be one of the following: {dummy}")
        self._Type = evap_type

    def __init__(self, ID='', ins=None, outs=(), *,
                 component='Water', P, V):
        Unit.__init__(self, ID, ins, outs)
        # Unpack
        out_wt_solids, liq = self.outs
        
        # Create components
        self._N_evap = n = len(P) # Number of evaporators
        Stream.species = liq.species
        evap0 = Evaporator_PV(None, outs=(None, None),
                             component=component, P=P[0])
        
        evaporators = [evap0]
        for i in range(1, n):
            evap = Evaporator_PQ(None,
                                   outs=(None, None, None),
                                   component=component, P=P[i], Q=0)
            evaporators.append(evap)
        condenser = HXutility(None, outs=Stream(None), V=0)
        evap0._heat_utilities[0], condenser._heat_utilities[0] = self._heat_utilities
        mixer = Mixer(None, outs=Stream(None))
        
        def V_error(v1):
            # Run first evaporator
            v_test = v1
            evap0.V = v1
            evap0._run()
            # Put liquid first, then vapor side stream
            ins = [evap0.outs[1], evap0.outs[0]]
            for i in range(1, n):
                evap = evaporators[i]
                evap._ins[:] = ins
                evap._run()
                v_test += (1-v_test) * evap._V
                # Put liquid first, then vapor side stream
                ins = [evap.outs[1], evap.outs[0]]
            return V - v_test
        self._V_error = V_error
        self.components = {'evaporators': evaporators,
                           'condenser': condenser,
                           'mixer': mixer}
        

    def _run(self):
        out_wt_solids, liq = self.outs
        ins = self.ins

        n = self._N_evap  # Number of evaporators

        # Set-up components
        components = self.components
        evaporators = components['evaporators']
        evaporators[0].ins[:] = [Stream.like(i, None) for i in ins]
        condenser = components['condenser']
        mixer = components['mixer']
        brentq(self._V_error, 0.0001, 0.9909, xtol=0.0001)
        
        # Condensing vapor from last effector
        
        outs_vap = evaporators[-1].outs[0]
        condenser.ins[:] = [outs_vap]
        condenser._run()
        outs_liq = [condenser.outs[0]]  # list containing all output liquids

        # Unpack other output streams
        out_wt_solids.copylike(evaporators[-1].outs[1])
        for i in range(1, n):
            evap = evaporators[i]
            outs_liq.append(evap.outs[2])

        # Mix liquid streams
        mixer.ins[:] = outs_liq
        mixer._run()
        liq.copylike(mixer.outs[0])
        
    def _design(self):
        # This functions also finds the cost
        A_range, C_func, U, _ = self._evap_data
        components = self.components
        evaporators = components['evaporators']
        Design = self._Design
        Cost = self._Cost
        CE = bst.CE
        
        evap0 = evaporators[0]
        hu = evap0._heat_utilities[0]
        duty = evap0._H_out - evap0._H_in
        hu(duty, evap0.ins[0].T, evap0.outs[0].T)
        Q = abs(duty)
        Tci = evap0.ins[0].T
        Tco = evap0.outs[0].T
        Th = evap0._heat_utilities[0]._fresh.T
        LMTD = ht.LMTD(Th, Th, Tci, Tco)
        ft = 1
        A = HXutility._calc_area(LMTD, U, Q, ft)
        self._evap_costs = evap_costs = [C_func(A, CE)]
        
        # Find condenser requirements
        condenser = components['condenser']
        condenser._design()
        condenser._cost()
        Cost['Condenser'] = condenser._Cost['Heat exchanger']
        
        # Find area and cost of evaporators
        As = [A]
        for evap in evaporators[1:]:
            Q = evap._Q
            Tc = evap.outs[0].T
            Th = evap.outs[2].T
            LMTD = Th - Tc
            A = HXutility._calc_area(LMTD, U, Q, ft)
            As.append(A)
            if not A_range[0] < A < A_range[1]:
                warn('area requirement ({A}) is out of range, {A_range}')
            evap_costs.append(C_func(A, CE))
        self._As = As
        Design['Area'] = A = sum(As)
        Design['Volume'] = vol = self._N_evap * self.tau * self.ins[0].volnet
        Cost['Evaporators'] = sum(evap_costs)
        
        # Calculate power
        power, cost = vacuum_system(massflow=0, volflow=0,
                                    P_suction=evap.outs[0].P, vol=vol,
                                    vacuum_system_preference='Liquid-ring pump')
        Cost['Vacuum liquid-ring pump'] = cost
        self._power_utility(power)
        
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 21 11:05:24 2017

The oil and sugar separation (pretreatment) section for the baseline lipid cane biorefinery is defined here as System objects. The systems include all streams and units starting from enzyme treatment to purification of the sugar solution and the oil stream.

@author: Yoel
"""
import numpy as np
from biosteam import System, Stream
from biosteam.units import Mixer, EnzymeTreatment, CrushingMill, \
    HXutility, RVF, SplitFlash, VibratingScreen, \
    MagneticSeparator, Clarifier, MixTank, \
    Shredder, ConveyingBelt, Splitter, \
    SplitCentrifuge_LLE, Pump, StorageTank
from biosteam.biorefineries.lipidcane.species import pretreatment_species
from biosteam.biorefineries.lipidcane.process_settings import price

__all__ = ('pretreatment_sys', 'lipid_cane', 'area_100', 'area_200')

# %% Species

Stream.species = pretreatment_species
psp = ('Ash', 'CaO', 'Cellulose', 'Ethanol', 'Flocculant',
       'Glucose', 'Hemicellulose', 'Lignin', 'Lipid',
       'Solids', 'H3PO4', 'Sucrose', 'Water')
psp1 = ('Ash', 'Cellulose', 'Glucose', 'Hemicellulose',
        'Lignin', 'Lipid', 'Solids', 'Sucrose', 'Water')
psp2 = ('Ash', 'CaO', 'Cellulose', 'Flocculant', 'Glucose',
        'Hemicellulose', 'Lignin', 'Lipid',
        'H3PO4', 'Sucrose', 'Water')

# %% Streams

f1 = (2000.042, 26986.69 , 2007.067, 15922.734, 14459.241,
      10035.334, 5017.667, 22746.761, 234157.798)
lipid_cane = Stream('lipid_cane', f1, psp1, units='kg/hr',
                    price=price['Lipid cane'])

enzyme = Stream('enzyme', Cellulose=100, Water=900, units='kg/hr',
                price=price['Protease'])

imbibition_water = Stream('imbibition_water',
                          Water=87023.35,
                          T = 338.15, units='kg/hr')

H3PO4 = Stream('H3PO4', H3PO4=74.23, Water=13.10, units='kg/hr',
               price=price['H3PO4'])  # to T203

lime = Stream('lime', CaO=333.00, Water=2200.00, units='kg/hr',
              price=price['Lime'])  # to P5

polymer = Stream('polymer', Flocculant=0.83, units='kg/hr',
                 price=price['Polymer'])  # to T205

rvf_wash_water = Stream('rvf_wash_water',
                        Water=16770, units='kg/hr',
                        T=363.15)  # to C202

oil_wash_water = Stream('oil_wash_water',
                        Water=1350, units='kg/hr',
                        T=358.15)  # to T207

# %% Units

Stream.default_ID = 'd'
Stream.default_ID_number = 0
# Stream.default_ID_number = 100

# Feed the shredder
U101 = ConveyingBelt('U101', ins=lipid_cane)
U101.cost_items['Conveying belt'].ub = 2500

# Separate metals
U102 = MagneticSeparator('U102', ins=U101.outs)

# Shredded cane
U103 = Shredder('U103', ins=U102.outs)

# Stream.default_ID_number = 200

# Hydrolyze starch
T201 = EnzymeTreatment('T201', T=323.15)  # T=50

# Finely crush lipid cane
U201 = CrushingMill('U201',
                    split=(0.92, 0.92, 0.04, 0.92, 0.92, 0.04, 0.1, 1),
                    order=('Ash', 'Cellulose', 'Glucose', 'Hemicellulose',
                           'Lignin', 'Sucrose', 'Lipid', 'Solids'),
                    moisture_content=0.5)

# Convey out bagasse
U202 = ConveyingBelt('U202', ins=U201.outs[0], outs='Bagasse')

# Mix in water
M201 = Mixer('M201')

# Screen out fibers
S201 = VibratingScreen('S201',
                       split=(0.35, 0.35, 0.88, 0.35,
                              0.35, 0.88, 0, 0.88, 0.88),
                       order=psp1)

# Store juice before treatment
T202 = StorageTank('T202')
T202.tau = 12

# Heat up before adding acid
H201 = HXutility('H201', T=343.15)

# Mix in acid
T203 = MixTank('T203')

# Pump acid solution
P201 = Pump('P201')

# Mix lime solution
T204 = MixTank('T204')
T204.tau = 1
P202 = Pump('P202')

# Blend acid lipid solution with lime
T205 = MixTank('T205')

# Mix recycle
M202 = Mixer('M202')

# Heat before adding flocculant
H202 = HXutility('H202', T=372.15)

# Mix in flocculant
T206 = MixTank('T206')
T206.tau = 1/4

# Separate residual solids
C201 = Clarifier('C201',
                split=(0, 0, 0, 0.522, 0.522, 0, 0,
                       0.98, 0.522, 0.522, 0.522),
                order=psp2)

# Remove solids as filter cake
C202 = RVF('C202', 
           outs=('filte_cake', ''),
           moisture_content=0.80,
           split=(0.85, 0.85, 0.85, 0.01, 0.85, 0.85, 0.01),
           order=('Ash', 'CaO', 'Cellulose', 'Glucose',
                  'Hemicellulose', 'Lignin', 'Sucrose'))
P203 = Pump('P203')

# Separate oil and sugar
T207 = MixTank('T207', outs=('', ''))
split = np.zeros(len(pretreatment_species), float)
index = pretreatment_species.indices(('Lipid', 'Water'))
split[index] = (1, 0.0001)
T207._split = split
T207._run = lambda : Splitter._run(T207)
del split, index

# Cool the oil
H203 = HXutility('H203', T=343.15)

# Screen out small fibers from sugar stream
S202 = VibratingScreen('S202', outs=('', 'fiber_fines'),
                      split=1-np.array((0, 0, 0, 1, 0.002, 0, 0,0, 0, 0.002, 0.002)),
                      order=psp2)
sugar = S202-0
S202.mesh_opening = 2

# Add distilled water to wash lipid
T208 = MixTank('T208')
T208.tau = 2

# Centrifuge out water
C203 = SplitCentrifuge_LLE('C203',
                           split=(0.99, 0.01),
                           order=('Lipid', 'Water'))

# Vacume out water
F201 = SplitFlash('F201', T=347.15, P=2026.5,
                 split=(0.0001, 0.999), order=('Lipid', 'Water'))
lipid = F201.outs[1]

# %% Process specifications

# Specifications dependent on lipid cane flow rate
_enzyme_mass = enzyme.mass[[9, 12]]
_CaO_Water_mass = lime.mass[[7, 12]]
_H3PO4_Water_mass = H3PO4.mass[[1, 12]]
last_lipidcane_massnet = int(lipid_cane.massnet)
def correct_flows():
    global last_lipidcane_massnet
    massnet = lipid_cane.massnet
    if int(massnet) != last_lipidcane_massnet:
        # correct enzyme, lime, phosphoric acid, and imbibition water
        _enzyme_mass[:] = 0.003 * massnet * np.array([0.1, 0.9])
        _CaO_Water_mass[:] = 0.001 * massnet * np.array([0.046, 0.954])
        _H3PO4_Water_mass[:] = 0.00025 * massnet
        imbibition_water_mass.value = 0.25* massnet
        last_lipidcane_massnet = int(massnet)

# Specifications within a system
def correct_lipid_wash_water():
    oil_wash_water.mol[12] = H202.outs[0].mol[-2]*100/11

solids_index = Stream.indices(['Ash', 'CaO', 'Cellulose', 'Hemicellulose', 'Lignin'])
def correct_wash_water():
    solids = solidsmol[solids_index].sum()
    rvf_wash_water.mol[12] = 0.0574*solids

imbibition_water_mass = imbibition_water.mass.item(12)


# %% Pretreatment system set-up

(U103-0, enzyme)-T201
(T201-0, M201-0)-U201-1-S201-0-T202
(S201-1, imbibition_water)-M201
crushing_mill_recycle_sys = System('crushing_mill_recycle_sys',
                               network=(U201, S201, M201),
                               recycle=M201-0)

T202-0-H201
(H201-0, H3PO4)-T203-P201
(P201-0, lime-T204-0)-T205-P202
(P202-0, P203-0)-M202-H202
(H202-0, polymer)-T206-C201
(C201-1, rvf_wash_water)-C202-1-P203
clarification_recycle_sys = System('clarification_recycle_sys',
                                   network=(M202, H202, T206, C201, C202, P203),
                                   recycle=C202-1)

C201-0-T207-0-H203
(H203-0, oil_wash_water)-T208-C203-0-F201
T207-1-S202

pretreatment_sys = System('pretreatment_sys',
                          network=(U101, U102, U103,
                                   correct_flows, T201,
                                   crushing_mill_recycle_sys,
                                   U202, T202, H201, T203,
                                   P201, T204, T205, P202,
                                   correct_wash_water,
                                   clarification_recycle_sys,
                                   T207, H203, S202,
                                   correct_lipid_wash_water,
                                   T208, C203, F201,))

solidsmol = P202.outs[0].mol

area_100 = System('area_100', network=(U101, U102, U103))
units = pretreatment_sys.units.copy()
for i in area_100.network: units.discard(i)
area_200_network = sorted(units, key=lambda x: x.ID)
area_200 = System('area_200', network=area_200_network)



# -*- coding: utf-8 -*-
"""
Created on Wed May  1 19:05:53 2019

@author: yoelr
"""
import biosteam as bst
import copy
from ._design import design
from math import ceil

__all__ = ('cost', 'add_cost', 'CostItem')

class CostItem:
    """Create a CostItem object which defines exponential scaling factors for an item's purchase cost.
    
    Parameters
    ----------
    basis : str
        Name of size parameter used for scaling.
    units : str
        Units of measure.
    S : float
        Size.
    ub : float
         Size limit.
    CE : float
         Chemical engineering plant cost index.
    cost : float
        Purchase cost of item.
    n : float
        Exponential factor.
    kW : float
        Electricity rate.
    BM : float
        Bare module factor (installation factor).
    N : str
        Attribute name for number of parallel units.
    """
    __slots__ = ('_basis', '_units', 'S', 'ub', 'CE',
                 'cost', 'n', 'kW', 'BM', 'N')
    def __init__(self, basis, units, S, ub, CE, cost, n, kW, BM, N):
        if ub and not N: N='#'
        self._basis = basis
        self._units = units
        self.S = S
        self.ub = ub
        self.CE = CE
        self.cost = cost
        self.n = n
        self.kW = kW
        self.BM = BM
        self.N = N
    
    __getitem__ = object.__getattribute__
    __setitem__ = object.__setattr__
    
    @property
    def basis(self):
        return self._basis
    @property
    def units(self):
        return self._units
    
    def __repr__(self):
        return f"<{type(self).__name__}: {self._basis} ({self._units})>"
    
    def _ipython_display_(self):
        print(f"{type(self).__name__}: {self._basis} ({self._units})\n"
             +f" S     {self.S:.3g}\n"
             +f" ub    {self.ub:.3g}\n"
             +f" CE    {self.CE:.3g}\n"
             +f" cost  {self.cost:.3g}\n"
             +f" n     {self.n:.3g}\n"
             +f" kW    {self.kW:.3g}\n"
             +f" BM    {self.BM:.3g}\n"
            +(f" N     '{self.N}'" if self.N else ""))
    show = _ipython_display_

def _cost(self):
    D = self._Design
    C = self._Cost
    kW = 0
    for i, x in self.cost_items.items():
        S = D[x._basis]
        if x.ub:
            D[x.N] = N = ceil(S/x.ub)
            q = S/x.S
            F = q/N
            C[i] = N*bst.CE/x.CE*x.cost*F**x.n
            kW += x.kW*q
        elif x.N:
            N = getattr(self, x.N)
            F = S/x.S
            C[i] = N*bst.CE/x.CE*x.cost*F**x.n
            kW += N*x.kW*F
        else:
            F = S/x.S
            C[i] = bst.CE/x.CE*x.cost*F**x.n
            kW += x.kW*F
    if kW: self._power_utility(kW)

def _extended_cost(self):
    _cost(self)
    self._end_decorated_cost_()

@property
def BM(self):
    raise AttributeError("can only get installation factor 'BM' through 'cost_items'")
@BM.setter 
def BM(self, BM):
    raise AttributeError("can only set installation factor 'BM' through 'cost_items'")
BM_property = BM
del BM

@property
def installation_cost(self):
    C = self._Cost
    return sum([C[i]*j.BM for i,j in self.cost_items.items()])

def cost(basis, ID=None, *, CE, cost, n, S=1, ub=0, kW=0, BM=1, units=None, fsize=None, N=None):    
    r"""Add item (free-on-board) purchase cost based on exponential scale up:
    
    Parameters
    ----------
    basis : str
        Name of size parameter used for scaling.    
    ID : str
        Name of purchase item.
    CE : float
         Chemical engineering plant cost index.
    cost : float
        Purchase cost of item.
    n : float
        Exponential factor.
    S = 1 : float
        Size.
    ub : float
         Size limit.
    kW : float
        Electricity rate.
    BM = 1: float
        Bare module factor (installation factor).
    units = None : str, optional
        Units of measure.
    fsize = None : function, optional
        Accepts a Unit object argument and returns the size parameter. If None, defaults to function predefined for given name and units of measure.
    N : str
        Attribute name for number of parallel units.
        
    Examples
    --------
    :doc:`../../tutorial/Unit decorators`
    
    """
    
    return lambda cls: add_cost(cls, ID, basis, units, S, ub, CE, cost, n, kW, BM, fsize, N)

def add_cost(cls, ID, basis, units, S, ub, CE, cost, n, kW, BM, fsize, N):
    if kW: cls._has_power_utility = True
    if basis in cls._units:
        if fsize:
            raise RuntimeError(f"cost basis '{basis}' already defined in class, cannot pass 'fsize' argument")
        elif not units: 
            units = cls._units[basis]
        elif units != cls._units[basis]:
            raise RuntimeError(f"cost basis '{basis}' already defined in class with units '{cls._units[basis]}'")
    elif units:
        design._add_design2cls(cls, basis, units, fsize)
    else:
        raise RuntimeError(f"units of cost basis '{basis}' not available in '{cls.__name__}._units' dictionary, must pass units or define in class")
    if hasattr(cls, 'cost_items'):
        if 'cost_items' not in cls.__dict__:
            cls.cost_items = copy.deepcopy(cls.cost_items)
        if not ID:
            raise ValueError("must pass an 'ID' for purchase cost item")
        if ID in cls.cost_items:
            raise ValueError(f"ID '{ID}' already in use")
        cls.cost_items[ID] = CostItem(basis, units, S, ub, CE, cost, n, kW, BM, N)
    else:
        ID = ID or cls.line
        cls.cost_items = {ID: CostItem(basis, units, S, ub, CE, cost, n, kW, BM, N)}
        if '_cost' not in cls.__dict__:
            if hasattr(cls, '_end_decorated_cost_'):
                cls._cost = _extended_cost
            else:
                cls._cost = _cost
        cls.BM = BM_property
        cls.installation_cost = installation_cost
    return cls




# -*- coding: utf-8 -*-
"""
Created on Sun Jul 22 12:59:22 2018

This lipid specie is defined here as tripalmitin.

@author: Yoel Rene Cortes-Pena
"""
import numpy as np
import pandas as pd
from biosteam.compounds import Liquid
import os

__all__ = ['lipid']

MW = 885.432
const_props = {'MW': MW,
               'sigma': 34.7*0.001,  # ChemCad 77F, glyceryl palmitate
               'Tc': 954.1,
               'omega': 1.68620002269745,  # http://www.ethermo.us/Mars1803.htm
               'dipole': 3.14782166481018,  # http://www.ethermo.us/Mars1803.htm
               'Hf': -1844000}  # http://www.ethermo.us/Mars1803.htm
dep_props = {}

# Load material properties
dir_path = os.path.dirname(os.path.realpath(__file__))
data_frame = pd.read_excel(dir_path + '/tripalmitin_liquid.xlsx')
T, rho, mu, nu, Cp, k = np.array(data_frame).transpose()

# Density
density = np.poly1d(np.polyfit(T, rho, 4))
dep_props['rho'] = lambda self: density(self.T)
molar_density = np.poly1d(np.polyfit(T, rho/MW, 4))
dep_props['rhom'] = lambda self: molar_density(self.T)
molar_vol = np.poly1d(np.polyfit(T, MW/rho/1000, 4))
dep_props['Vm'] = lambda self: molar_vol(self.T)

# Dynamic viscosity
dviscosity = np.poly1d(np.polyfit(T, mu, 4))
dep_props['mu'] = lambda self: dviscosity(self.T)

# Kinematic viscosity
kviscosity = np.poly1d(np.polyfit(T, nu, 4))
dep_props['nu'] = lambda self: kviscosity(self.T)

# Heat capacity
molar_heat_capacity = np.poly1d(np.polyfit(T, Cp*MW, 4))
dep_props['Cpm'] = lambda self: molar_heat_capacity(self.T)  # J/mol/K
heat_capacity = np.poly1d(np.polyfit(T, Cp, 4))
dep_props['Cp'] = lambda self: heat_capacity(self.T)  # kJ/kg/K

# Thermal conductivity
conductivity = np.poly1d(np.polyfit(T, k, 4))
dep_props['k'] = lambda self: conductivity(self.T)

# Prandtl number
dep_props['Pr'] = lambda self: self.Cp*self.mu/self.k*1000

# Tripalmitin class
Lipid = Liquid.Factory('Lipid', const_props, dep_props)

# Make the new chemical object
lipid = Lipid('Tripalmitin')
for i, j in const_props.items():
    lipid.__dict__[i] = j
lipid.VaporPressure = lambda T: np.exp(6.596 - 5407.33/T)
lipid.phase_ref = 'l'





# -*- coding: utf-8 -*-
"""
Created on Mon Sep 16 00:02:57 2019

@author: yoelr
"""
import pandas as pd
from biosteam.evaluation.evaluation_tools import plot_spearman

# %%  Plot Spearman correlations

# Replacement parameter labels
replacement_labels = {
    'Stream-Ethanol price': 'Ethanol price',
    'TEA operating days': 'Operating days',
    'Stream-Cornstover price': 'Cornstover price',
    'Fermentation-U34 efficiency': 'Fermentation efficiency',
    'Stream-Cellulase price': 'Cellulase price',
    'Stream-Cornstover flow rate': 'Cornstover flow rate',
    'TEA income tax': 'Income tax',
    'Saccharification and co fermentation-U72 saccharification conversion': 'Saccharification conversion',
    'Saccharification and co fermentation-U72 ethanol conversion': 'Ethanol conversion',
    'Boiler turbogenerator-BT boiler efficiency': 'Boiler efficiency',
    'Boiler turbogenerator boiler base cost': 'Boiler base cost',
    'Boiler turbogenerator turbogenerator base cost': 'Turbogenerator base cost',
    'Pretreatment reactor system base cost': 'Pretreatment reactor base cost',
    'Power utility price': 'Electricity price',
    'Cooling tower base cost': 'Cooling tower base cost',
    'Waste water system cost waste water system base cost': 'Wastewater treatment base cost'}


def replace_label_text(label_text):
    """Replace label text for graph."""
    name, distribution = label_text.split(' (')
    lb, mid, ub = eval('(' + distribution)
    if 'efficiency' in name:
        distribution = f" ({lb:.2f}, {mid:.2f}, {ub:.2f})"
    else:
        distribution = f" ({lb:.3g}, {mid:.3g}, {ub:.3g})"
    pos = name.find(' [')
    if pos != -1:
        units = name[pos:]
        name = name[:pos]
    else:
        units = ''
    if name in replacement_labels:
        name = replacement_labels[name]
    return name + units + distribution

# Get data
rhos = pd.read_excel('Spearman correlation cornstover.xlsx',
                     header=[0], dtype=float)['Minimum ethanol selling price']

# Get only important parameters
rhos = rhos[rhos.abs()>0.055] 

# Plot and fix axis labels
fig, ax = plot_spearman(rhos, top=10)
labels = [item.get_text() for item in ax.get_yticklabels()]
new_labels = [replace_label_text(i) for i in labels]
ax.set_yticklabels(new_labels)


# -*- coding: utf-8 -*-
"""
Created on Thu Mar  7 23:17:38 2019

@author: yoelr
"""
import numpy as np
import pandas as pd
from ._state import State
from ._metric import Metric
from biosteam.utils import TicToc
from scipy.stats import spearmanr

__all__ = ('Model',)

# %% Functions

varindices = lambda vars: [var.index for var in vars]
varcolumns = lambda vars: pd.MultiIndex.from_tuples(
                                 varindices(vars),
                                 names=('Element', 'Variable'))

    
# %% Grid of simulation blocks

class Model(State):
    """Create a Model object that allows for evaluation over a sample space.
    
    Parameters
    ----------
    system : System
        Should reflect the model state.
    metrics : tuple[Metric]
        Metrics to be evaluated by model.
    skip=False : bool
        If True, skip simulation for repeated states.
    Examples
    --------
    :doc:`../tutorial/Advanced simulation`
    
    """
    __slots__ = ('_table',   # [DataFrame] All arguments and results
                 '_metrics', # tuple[Metric] Metrics to be evaluated by model
                 '_index',   # list[int] Order of sample evaluation for performance
                 '_samples', # [array] Argument sample space
                 '_setters') # list[function] Cached parameter setters
    
    def __init__(self, system, metrics, skip=False):
        super().__init__(system, skip)
        self.metrics = metrics
        self._setters = self._samples = self._table = None
        
    def copy(self):
        """Return copy."""
        copy = super().copy()
        copy._metrics = self._metrics
        if self._update:
            copy._setters = self._setters
            copy._table = self._table.copy()
        else:
            copy._setters = copy._samples = copy._table = None
        return copy
    
    def _erase(self):
        """Erase cached data."""
        self._setters = self._update = self._table = self._samples = None
    
    @property
    def metrics(self):
        """tuple[Metric] Metrics to be evaluated by model."""
        return self._metrics
    @metrics.setter
    def metrics(self, metrics):
        for i in metrics:
            if not isinstance(i, Metric):
                raise ValueError(f"metrics must be '{Metric.__name__}' "
                                 f"objects, not '{type(i).__name__}'")
        self._metrics = tuple(metrics)
    
    @property
    def table(self):
        """[DataFrame] Table of the sample space with results in the final column."""
        return self._table
    
    def load_samples(self, samples):
        """Load samples for evaluation
        
        Parameters
        ----------
        samples : numpy.ndarray, dim=2
        
        """
        if not self._update: self._loadparams()
        params = self._params
        paramlen = len(params)
        if not isinstance(samples, np.ndarray):
            raise TypeError(f'samples must be an ndarray, not a {type(samples).__name__} object')
        elif samples.ndim != 2:
            raise ValueError('samples must be 2 dimensional')
        elif samples.shape[1] != paramlen:
            raise ValueError(f'number of parameters in samples ({samples.shape[1]}) must be equal to the number of parameters ({len(params)})')
        key = lambda x: samples[x][i]
        N_samples = len(samples)
        index = list(range(N_samples))
        for i in range(paramlen-1,  -1, -1):
            if not params[i].system: break
            index.sort(key=key)
        self._index = index
        empty_metric_data = np.zeros((N_samples, len(self.metrics)))
        self._table = pd.DataFrame(np.hstack((samples, empty_metric_data)),
                                   columns=varcolumns(tuple(params)+self.metrics))
        self._samples = samples
        
    def evaluate(self, thorough=True):
        """Evaluate metric over the argument sample space and save values to ``table``.
        
        Parameters
        ----------
        thorough : bool
            If True, simulate the whole system with each sample.
            If False, simulate only the affected parts of the system.
        
        """
        # Setup before simulation
        funcs = [i.getter for i in self._metrics]
        values = []
        add = values.append
        samples = self._samples
        if samples is None: raise ValueError('must load samples or distribution before evaluating')
        if thorough:
            if self._setters:
                setters = self._setters
            else:
                self._setters = setters = [p.setter for p in self._params]
            simulate = self._system.simulate
            zip_ = zip
            for i in self._index:
                for f, s in zip_(setters, samples[i]): f(s)
                simulate()
                add([i() for i in funcs])
        else:
            update = self._update
            for i in self._index: 
                update(samples[i])
                add([i() for i in funcs])    
        cols = varindices(self._metrics)
        for k, v in zip_(cols, zip_(*values)): self.table[k] = v
    
    def evaluate_across_coordinate(self, f_coordinate, name,
                                   lb, ub, xlfile,
                                   N_samples=1000, N_points=20,  rule='L',
                                   notify=True):
        """Evaluate across coordinate and save sample metrics.
        
        Parameters
        ----------
        f_coordinate : function
            Should change state of system given the coordinate.
        lb : float
            Lower bound of coordinate.
        ub : float
            Upper bound of coordinate.
        xlfile : str
            Name of file to save. File must end with ".xlsx"
        N_samples=1000 : int, optional
            Number of samples at each coordinate.
        N_points=20 : int, optional
            Number of lipid fractions evaluated.
        rule='L' : str
            Sampling rule.
        notify=True : bool, optional
            If True, notify elapsed time after each coordinate evaluation.
        
        """
        # Load samples
        samples = self.sample(N_samples, 'L')
        self.load_samples(samples)
        
        # Points to evaluate
        coordinate = np.linspace(lb, ub, N_points)
        
        # Initialize data containers
        metric_data = {}
        def new_data(key, dct=metric_data):
            data = np.zeros([N_samples, N_points])
            dct[key] = data
            return data
        for i in self.metrics: new_data(i.index)
        
        # Initialize timer
        if notify:
            timer = TicToc()
            timer.tic()
            def evaluate():
                self.evaluate()
                print(f"Elapsed time: {timer.elapsed_time:.0f} sec")
        else:
            evaluate = self.evaluate
        
        # Simulate lipidcane biorefinery    
        table = self.table
        for n, f in enumerate(coordinate):
            f_coordinate(f)
            evaluate()
            for metric in metric_data:
                metric_data[metric][:, n] = table[metric]
        
        # Save data to excel
        data = pd.DataFrame(data=np.zeros([N_samples, N_points]), columns=coordinate)
        data.columns.name = name
        
        with pd.ExcelWriter(xlfile) as writer:
            for metric in metric_data:
                data[:] = metric_data[metric]
                data.to_excel(writer, sheet_name=metric[-1])
    
    def spearman(self, metrics=(), excluded_params=()):
        """Return DataFrame of Spearman's rank correlation for metrics vs parameters.
        
        Parameters
        ----------
        metrics=() : Iterable[Metric], defaults to all metrics
            Names of metrics to be be correlated with parameters.
        excluded_params=() : Iterable[str], optional
            Names of parameters to exclude.
        """
        data = self.table
        param_cols = list(data)
        all_metric_names = varindices(self.metrics)
        if not metrics: 
            metric_names = all_metric_names
        else:
            metric_names = varindices(metrics)
        params = list(self._params)
        for i in excluded_params:
            for j in params:
                if j.name == i:
                    params.remove(j)
                    break
            for j in param_cols: 
                if j[0] == i:
                    param_cols.remove(j)
                    break
        
        for i in all_metric_names: param_cols.remove(i)
        param_descriptions = [i.describe() for i in params]
        allrhos = []
        for name in metric_names:
            rhos = []
            metric = data[name]
            for col in param_cols:
                rho, p = spearmanr(data[col], metric)
                rhos.append(rho)
            allrhos.append(rhos)
        allrhos = np.array(allrhos).transpose()
        return pd.DataFrame(allrhos, index=param_descriptions,
                            columns=metric_names)
    
    def __call__(self, sample):
        """Return list of metric values."""
        super().__call__(sample)
        return [i.getter() for i in self._metrics]
    
    def _repr(self):
        clsname = type(self).__name__
        newline = "\n" + " "*(len(clsname)+2)
        return f'{clsname}: {newline.join([i.describe() for i in self.metrics])}'
        
    def __repr__(self):
        return f'<{type(self).__name__}: {", ".join([i.name for i in self.metrics])}>'
    
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 16 11:49:52 2018

@author: Yoel Rene Cortes-Pena
"""
from . import array_utils
from . import color_utils
from . import other_utils
from . import stream_utils
from . import display_units
from . import plot_utils
from . import register
from . import tictoc
from . import solvers
from . import not_implemented_method
from . import network_utils

from .not_implemented_method import *
from .array_utils import *
from .color_utils import *
from .other_utils import *
from .stream_utils import *
from .display_units import *
from .plot_utils import *
from .tictoc import *
from .register import *
from .solvers import *
from .network_utils import *

__all__ = ['array_utils', 'color_utils',
           'other_utils', 'stream_utils',
           'display_utils', 'plot_utils',
           'register', 'tictoc', 'solvers',
           'not_implemented_method',
           'network_utils']
__all__.extend(plot_utils.__all__)
__all__.extend(not_implemented_method.__all__)
__all__.extend(solvers.__all__)
__all__.extend(display_units.__all__)
__all__.extend(array_utils.__all__)
__all__.extend(color_utils.__all__)
__all__.extend(other_utils.__all__)
__all__.extend(stream_utils.__all__)
__all__.extend(register.__all__)
__all__.extend(tictoc.__all__)
__all__.extend(network_utils.__all__)


# -*- coding: utf-8 -*-
"""
Created on Thu Aug  1 21:48:12 2019

@author: yoelr
"""

from biosteam import TEA

__slots__ = ('CornstoverTEA',)

class CornstoverTEA(TEA):
    
    __slots__ = ('OSBL_units', 'warehouse', 'site_development',
                 'additional_piping', 'proratable_costs', 'field_expenses',
                 'construction', 'contingency', 'other_indirect_costs', 
                 'labor_cost', 'labor_burden', 'property_insurance',
                 'maintenance', '_ISBL_DPI_cached', '_FCI_cached',
                 '_utility_cost_cached')
    
    def __init__(self, system, IRR, duration, depreciation, income_tax,
                 operating_days, lang_factor, construction_schedule,
                 startup_months, startup_FOCfrac, startup_VOCfrac,
                 startup_salesfrac, WC_over_FCI,  finance_interest,
                 finance_years, finance_fraction, OSBL_units, warehouse,
                 site_development, additional_piping, proratable_costs,
                 field_expenses, construction, contingency,
                 other_indirect_costs, labor_cost, labor_burden,
                 property_insurance, maintenance):
        super().__init__(system, IRR, duration, depreciation, income_tax,
                         operating_days, lang_factor, construction_schedule,
                         startup_months, startup_FOCfrac, startup_VOCfrac,
                         startup_salesfrac, WC_over_FCI,  finance_interest,
                         finance_years, finance_fraction)
        self.OSBL_units = OSBL_units
        self.warehouse = warehouse
        self.site_development = site_development
        self.additional_piping = additional_piping
        self.proratable_costs = proratable_costs
        self.field_expenses = field_expenses
        self.construction = construction
        self.contingency = contingency
        self.other_indirect_costs = other_indirect_costs
        self.labor_cost = labor_cost
        self.labor_burden = labor_burden
        self.property_insurance = property_insurance
        self.maintenance = maintenance
    
    @property
    def utility_cost(self):
        self._utility_cost_cached = utility_cost = super().utility_cost
        return utility_cost
    
    def _ISBL_DPI(self, DPI):
        """Direct permanent investment of units inside battery limits."""
        if self.lang_factor:
            self._ISBL_DPI_cached = DPI - sum([i.purchase_cost for i in self.OSBL_units]) * self.lang_factor
        else:
            self._ISBL_DPI_cached = DPI - sum([i.installation_cost for i in self.OSBL_units])
        return self._ISBL_DPI_cached
        
    def _TDC(self, DPI):
        return DPI + self._ISBL_DPI(DPI)*(self.warehouse
                                          + self.site_development
                                          + self.additional_piping)
    
    def _indirect_costs(self, TDC):
        return TDC*(self.proratable_costs + self.field_expenses
                    + self.construction + self.contingency
                    + self.other_indirect_costs)
    
    def _FCI(self, TDC):
        self._FCI_cached = FCI = TDC + self._indirect_costs(TDC)
        return FCI
    
    def _FOC(self, FCI):
        return (FCI * self.property_insurance
                + self._ISBL_DPI_cached * self.maintenance
                + self.labor_cost*(1+self.labor_burden))


# -*- coding: utf-8 -*-
"""
Created on Tue May 28 11:11:08 2019

@author: yoelr
"""

from biosteam import PowerUtility, HeatUtility, find, Flowsheet
import biosteam

__all__ = ('price',)

# %% Process settings

biosteam.CE = 567 # 2013
find.set_flowsheet(Flowsheet('sugarcane'))
PowerUtility.price = 0.065
_ha = HeatUtility.heating_agents['Low pressure steam']
_ha.efficiency = 0.90
_ha.T = 529.2
_ha.P = 44e5
_ha.price_kmol = 0.30626
_ha.Hvap = 30235.86
HeatUtility.cooling_agents['Cooling water'].price_kmol = 0
HeatUtility.cooling_agents['Chilled water'].price_kJ = 0


# Raw material price (USD/kg)
price = {'Sugar cane': 0.03455, # 70% m.c
         'Water': 0.000353,
         'HCl': 0.205,
         'Lime': 0.077,
         'H3PO4': 0,#0.700, # TODO: find price
         'NaOH':0.41,
         'Protease': 0.5,
         'Polymer': 0, # TODO: find price
         'Steam': 0.017,
         'Ethanol': 0.789,
         'Waste': -0.33,
         'Gasoline': 0.756} # 2 USD/gal


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 15 09:36:15 2018

All species for the biodiesel production section of the lipid cane baseline biorefinery are defined here in the Species object, biodiesel_species.

@author: Yoel Rene Cortes-Pena
"""
from biosteam import Species
from biosteam.compounds import Chemical, Substance
from .lipid import lipid

__all__ = ['biodiesel_species']

# %% Make species

sp = Species('Methanol', 'Glycerol', 'Water')
sp.Lipid = lipid # Triolein
sp.Biodiesel = Chemical('methyl oleate')
# sp.FreeLipid = Chemical('oleic acid')

# %% Enthalpy of formations
sp.Lipid.Hf = -2193.7 * 1000 # https://webbook.nist.gov/cgi/cbook.cgi?ID=C122327&Mask=2
#sp.free_lipid.Hfm = -764.80 * 1000
sp.Biodiesel.Hf = -727.64 * 1000
sp.Glycerol.Hf = -669.6 * 1000

# %% Missing properties

sp.Lipid.Tb = 879.9 # Boiling point of Triolein
sp.Lipid.dipole = 0.1
sp.Biodiesel.dipole = 0

# %% Other solids

# Acid and base are basically water
HCl = Substance('HCl', MW=36.46094, rho=10**8)
NaOH = Substance('NaOH', MW=39.997109, rho=10**8)
NaOH.Hfm = 0

# Catalyst is basically methanol
NaOCH3 = Substance('NaOCH3', obj=sp.Methanol, MW=54.023689,
                         rho=1/sp.Methanol.Vm*54.03)
NaOCH3.Hfm = sp.Methanol.Hfm

biodiesel_species = Species.tospecies(tuple(sp) + (HCl, NaOH, NaOCH3))

# Set working species (and in this order)
biodiesel_species.IDs = ('Lipid', 'Methanol', 'Glycerol', 'Biodiesel',
                        'Water', 'NaOH', 'HCl', 'NaOCH3')


# -*- coding: utf-8 -*-
"""
Created on Tue Aug 13 21:44:23 2019

@author: yoelr
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from biosteam import colors
from biosteam.evaluation.evaluation_tools import plot_single_points, \
                                                 plot_montecarlo

light_color = colors.blue_tint.RGBn
dark_color = colors.blue_shade.RGBn
dot_color = colors.purple_shade.RGBn
nums = tuple(range(1, 9))
    
# %% Plot MESP

positions = (0,)
data = pd.read_excel('Monte Carlo cornstover.xlsx', header=[0, 1])
MESP = np.array(data[('Biorefinery', 'Minimum ethanol selling price')])
bx = plot_montecarlo(MESP, light_color, dark_color, positions, transpose=True)
sc = plot_single_points(positions, (2.15,), dot_color)

plt.ylim(1.65, 2.65)
plt.xlim(-1, 1)
plt.xticks([0], ['Cornstover'])
plt.ylabel('MESP ' + '[$\mathrm{USD} \cdot \mathrm{gal}^{-1}$]')
plt.tight_layout()
plt.legend([bx, sc], ['BioSTEAM', 'ASPEN (Humbird 2011)'])

# %% Plot electricity

areas = [f'Area {i}00' for i in nums]
xmarks = [i.replace(' ', '\n') for i in areas]
plt.figure()
units = 'MW'
positions = np.arange(0, 9)
electricity_cols = [(i, 'Electricity') for i in areas] + [('Biorefinery', 'Excess electricity')]
humbird_electricity = 41 * np.array((0.02, 0.14, 0.06, 0.05,
                                     0.18, 0.003, 0.03, 0.08, 0.44))
electricity_data = data[electricity_cols]  #/ humbird_electricity
electricity_data[('Biorefinery', 'Excess electricity')] /= 1000
plot_montecarlo(electricity_data, light_color, dark_color, transpose=True)
plot_single_points(positions, humbird_electricity, dot_color)
plt.xticks(positions, xmarks + ['Excess'])

# %% Plot installation cost

plt.figure()
units = '10^6 USD'
installation_cols = [(i, 'Installation cost') for i in areas[1:]]
humbird_installation = np.array([24.2, 32.9, 31.2, 22.3, 49.4, 5, 66, 6.9])
installation_data = data[installation_cols] / 1e6
plot_montecarlo(installation_data, light_color, dark_color, transpose=True)
plot_single_points(positions[:-2], humbird_installation[1:], dot_color)
plt.xticks(positions[:-1], xmarks[1:])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb  2 22:35:30 2018

All species for the ethanol production section of the lipid cane baseline biorefinery are defined here in the Species object, ethanol_species.

@author: Yoel Rene Cortes-Pena
"""
from biosteam import Species
from biosteam.compounds import Chemical, Gas, StaticChemical, Substance

__all__ = ['ethanol_species']

#%% Set up species

Water = Chemical('Water')        
CO2 = Gas('CO2')
Ethanol = Chemical('Ethanol')
_glucosoe = Chemical('Glucose')
glucose = StaticChemical('Glucose')
sucrose = StaticChemical('Sucrose')
dry_yeast = Substance('DryYeast', 'Yeast', MW=1)
phosphoric_acid = Substance('H3PO4', '7664-38-2', MW=97.994, rho=10**8)
Octane = Chemical('Octane')
ethanol_species = sp = Species.tospecies((CO2, Ethanol, Water, glucose, sucrose, phosphoric_acid, Octane, dry_yeast))

#%% Cp Gluclose & sucrose

# # Asadi, 2005, Tables, Beet-Sugar Handbook, John Wiley & Sons, Inc (2005), pp.  779-801
# # http://onlinelibrary.wiley.com/store/10.1002/9780471790990.oth1/asset/oth1.pdf;jsessionid=1B4B5D3655477F46D578541BE0E4CED2.f03t01?v=1&t=jdewgnrl&s=477661e0dfe5c058191f8f1907d2f63bdaca3e59

# # Use heat capacity at T = 60, and saturated with sugar
# sucrose.Cp = glucose.Cp = ((2540/4184 - 0.20)*4184 /0.8) * 180.156/1000**2  # J/mol/K
# sucrose.Cpm = sucrose.Cp * sucrose.MW
# glucose.Cpm = glucose.Cp * glucose.MW
dry_yeast.Cp = dry_yeast.Cpm = 1.2

#%% Cp Phophoric_acid

# https://pubs.acs.org/doi/pdf/10.1021/j150567a017
# HEAT CAPACITY OF PHOSPHORIC ACID SOLUTIONS, 15 TO 80'

# Use heat capacity at T = 59.749 C for 84.81% solution in Water
phosphoric_acid.Cp = (4.184*(0.4624 - 0.1519)/0.8481) * 97.994/1000  # J/mol/K
phosphoric_acid.Cpm = phosphoric_acid.Cp * phosphoric_acid.MW

#%% Enthalpy of formation (all information from NIST)

# The heats of formation (kJ/kmol)
Ethanol.Hf = -277.690 * 1000
Water.Hf = -241.820 * 1000
CO2.Hf = -393.520 * 1000  # in gas
glucose.Hf = -1274*1000
sucrose.Hf = -2221.2*1000
Octane.Hf = -250*1000
phosphoric_acid.Hf = -1271.66*1000
dry_yeast.Hf = 0



# -*- coding: utf-8 -*-
"""
Created on Sun Nov 11 11:20:42 2018

@author: yoelr
"""
from . import _Q
from .utils import DisplayUnits

__all__ = ('PowerUtility',)

class PowerUtility:
    """Create an PowerUtility object that, when called, calculates the cost of power (kW) and saves the power and cost.
            
    Examples
    --------
    Create a PowerUtility object:
    
    .. code-block:: python
    
       >>> pu = PowerUtility()
       >>> pu
       <PowerUtility: 0 kW, 0 USD/hr>
       
    Call object to calculate cost:
        
    .. code-block:: python
    
       >>> pu(rate=500)
       >>> pu
       <PowerUtility: 500 kW, 30 USD/hr>
       
    Results are accessible:
        
    .. code-block:: python
    
       >>> pu.rate, pu.cost
       (500, 30.)
    
    """
    _units = dict(rate='kW', cost='USD/hr')
    
    #: [DisplayUnits] Units of measure for IPython display
    display_units = DisplayUnits(**_units)
    
    __slots__ = ('rate', 'cost')
    #: [float] USD/kWhr
    price = 0.0782
    
    def __init__(self):
        #: Power requirement (kW)
        self.rate = 0
        
        #: Cost (USD/hr)
        self.cost = 0
    
    def __call__(self, rate):
        """Calculate cost and save. 
        
        Parameters
        ----------
        rate : float
               Power requirement (kW)
        
        """
        self.rate = rate
        self.cost = self.price * rate
    
    def __repr__(self):
        # Get units of measure
        units = self._units
        rate_units, cost_units = self.display_units
        rate = _Q(self.rate, units['rate']).to(rate_units).magnitude
        cost = _Q(self.cost, units['cost']).to(cost_units).magnitude
        return (f'<{type(self).__name__}: {rate:.3g} {rate_units}, {cost:.3g} {cost_units}>')
        
        


# -*- coding: utf-8 -*-
"""
Created on Mon Jul 15 09:20:57 2019

@author: yoelr
"""
from .. import Unit

class VentTank:
        
    def _run(self):
        feed = self.ins[0]
        vent, effluent = self.outs
        effluent.copylike(feed)
        P_vapor = effluent.P_vapor
        
        
        


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:17:05 2018

@author: yoelr
"""
from ._splitter import Splitter
from .decorators import cost

__all__ = ('MolecularSieve',)


# @cost('Flow rate', 'Pressure filter drying (2)',
#       cost=405000, CE=521.9, S=22687, n=0.6, kW=1044)
# @cost('Flow rate', 'Pressure filter pressing',
#       cost=75200, CE=521.9, S=22687, n=0.6, kW=112)
@cost('Flow rate', 'Column', kW=151, BM=1.8,
      cost=2601000, CE=521.9, S=22687, n=0.6)
class MolecularSieve(Splitter):
    """Create an ethanol/water molecular sieve for bioethanol plants.
    The molecular sieve is modeled as a component wise separator. Costing
    is based on scaling by the 6/10ths rule from an NREL TEA report [1].
    
    Parameters
    ----------
    ins
        [0] Feed (gas)
        
    outs
        [0] Split stream (gas)
        
        [1] Remainder stream (gas)
    split : array_like
            Componentwise split to the 0th output stream
    
    References
    ----------
    [1] Process Design and Economics for Biochemical Conversion of
    Lignocellulosic Biomass to Ethanol Dilute-Acid Pretreatment and
    Enzymatic Hydrolysis of Corn Stover. D. Humbird, R. Davis, L.
    Tao, C. Kinchin, D. Hsu, and A. Aden (National Renewable Energy
    Laboratory Golden, Colorado). P. Schoen, J. Lukas, B. Olthof,
    M. Worley, D. Sexton, and D. Dudgeon (Harris Group Inc. Seattle,
    Washington and Atlanta, Georgia)
    
    Examples
    --------
    :doc:`notebooks/MolecularSieve Example`
    
    """
    _N_heat_utilities = 2
    _units = {'Flow rate': 'kg/hr'}
    def __init__(self, ID='', ins=None, outs=(), *, order=None, split):
        Splitter.__init__(self, ID, ins, outs, order=order, split=split)
        self._outs[0]._phase = self._outs[1]._phase = 'g'

    def _design(self):
        self._Design['Flow rate'] = flow = self._outs[1].massnet
        T = self.ins[0].T
        self._heat_utilities[0](1429.65*flow, T)
        self._heat_utilities[1](-55.51*flow, T)




# -*- coding: utf-8 -*-
"""
Created on Tue Mar 26 14:45:00 2019

@author: yoelr
"""
__all__ = []

from .pretreatment import *
from .ethanol import *
from .biodiesel import *
from .biorefinery import *

from . import pretreatment
from . import ethanol
from . import biodiesel
from . import biorefinery

__all__.extend(pretreatment.__all__)
__all__.extend(ethanol.__all__)
__all__.extend(biodiesel.__all__)
__all__.extend(biorefinery.__all__)



# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 15:04:55 2018

@author: yoelr
"""
from scipy.optimize import newton
from ._exceptions import SolverError, _try_method
from ._flowsheet import find, make_digraph, save_digraph
from ._stream import Stream
from ._facility import Facility
from ._unit import Unit
from ._report import save_report
from .utils import colors, MissingStream, strtuple, \
                   conditional_wegstein, conditional_aitken, \
                   build_network
import biosteam as bst

__all__ = ('System',)

# %% Debugging and exception handling

def _evaluate(self, command=None):
    """Evaluate a command and request user input for next command. If no command, return. This function is used for debugging a System object."""    
    # Done evaluating if no command, exit debugger if 'exit'
    if command is None:
        Next = colors.next('Next: ') + f'{repr(self)}\n'
        info = colors.info("Enter to continue or type to evaluate:\n")
        command = input(Next + info + ">>> ")
    
    if command == 'exit': raise KeyboardInterrupt()
    if command:
        # Build locals dictionary for evaluating command
        lcs = {} 
        for attr in ('stream', 'unit', 'system'):
            dct = getattr(find, attr).__dict__
            lcs.update({i:j() for i, j in dct.items()})
        lcs.update({attr:getattr(bst, attr) for attr in bst.__all__})
        try:
            out = eval(command, {}, lcs)            
        except Exception as err:
            # Print exception and ask to raise error or continue evaluating
            err = colors.exception(f'{type(err).__name__}:') + f' {str(err)}\n\n'
            info = colors.info(f"Enter to raise error or type to evaluate:\n")
            command = input(err + info + ">>> ")
            if command == '': raise err
            _evaluate(self, command)        
        else:
            # If successful, continue evaluating
            if out is None: pass
            elif (not hasattr(out, '_ipython_display_')
                  or isinstance(out, type)): print(out)
            else: out._ipython_display_()
                
            command = input(">>> ")
            _evaluate(self, command)

def _method_debug(self, func):
    """Method decorator for debugging system."""
    def wrapper(*args, **kwargs):
        # Run method and ask to evaluate
        _evaluate(self)
        func(*args, **kwargs)
        
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    wrapper._original = func
    return wrapper

def _notify_run_wrapper(self, func):
    """Decorate a System run method to notify you after each loop"""
    def wrapper(*args, **kwargs):
        if self.recycle:
            func(*args, **kwargs)
            input(f'        Finished loop #{self._iter}\n')
        else:
            func(*args, **kwargs)
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    wrapper._original = func
    return wrapper
    
# %% System node for diagram

class _systemUnit(Unit, isabstract=True):
    """Dummy unit for displaying a system."""
    line = 'System'
    ID = None
    _N_ins = _N_outs = 1
    _ID = property(lambda self: self.ID)

_sysgraphics = _systemUnit._graphics
_sysgraphics.edge_in = _sysgraphics.edge_in * 10
_sysgraphics.edge_out = _sysgraphics.edge_out * 15
_sysgraphics.node['peripheries'] = '2'

class _streamUnit(Unit, isabstract=True):
    line = ''
    ID = None
    _ID = _systemUnit._ID
    
_stream_graphics = _streamUnit._graphics
_stream_graphics.node['fillcolor'] = 'white:#79dae8'
del _stream_graphics, _sysgraphics

# %% Other

_isfeed = lambda stream: not stream._source and stream._sink
_isproduct = lambda stream: not stream._sink and stream._source


# %% Process flow

class system(type):
    @property
    def converge_method(self):
        """Iterative convergence method ('wegstein', 'aitken', or 'fixed point')."""
        return self._converge.__name__[1:]

    @converge_method.setter
    def converge_method(self, method):
        method = method.lower().replace('-', '').replace(' ', '')
        if 'wegstein' == method:
            self._converge = self._wegstein
        elif 'fixedpoint' == method:
            self._converge = self._fixed_point
        elif 'aitken' == method:
            self._converge = self._aitken
        else:
            raise ValueError(f"only 'wegstein', 'aitken', and 'fixed point' methods are valid, not '{method}'")


class System(metaclass=system):
    """Create a System object that can iteratively run each element in a network of BioSTREAM objects until the recycle stream is converged. A network can have function, Unit and/or System objects. When the network contains an inner System object, it converges/solves it in each loop/iteration.

    Parameters
    ----------
    ID : str
         A unique identification. If ID is None, instance will not be
         registered in flowsheet.
    network : tuple[Unit, function and/or System]
        A network that is run element by element until the recycle converges.
    recycle=None : Stream, optional
        A tear stream for the recycle loop.
    facilities=() : tuple[Unit, function, and/or System], optional
        Offsite facilities that are simulated only after
        completing the network simulation.

    """
    ### Class attributes ###
    
    #: Maximum number of iterations
    maxiter = 100

    #: Molar tolerance (kmol/hr)
    molar_tolerance = 0.01
    
    #: Temperature tolerance (K)
    T_tolerance = 0.10

    # [dict] Cached downstream systems by (system, unit, with_facilities) keys
    _cached_downstream_systems = {} 

    @classmethod
    def from_feedstock(cls, ID, feedstock):
        return NotImplemented
        if not _isfeed(feedstock):
            raise ValueError('feedstock must have no upstream units')
        sink = feedstock.sink
        all_units = sink._neighborhood(10000)
        sources = [sink]
        for i in all_units:
            if any([i.source for i in i._ins]): continue
            sources.append(i)
        return build_network(sources)
            
    @classmethod
    def from_area(cls, ID, sources, sinks=()):
        return NotImplemented
        network = build_network(sources, sinks)
        system = cls(ID, network)
        return system
    
    def __init__(self, ID, network, recycle=None, facilities=()):
        
        #: Molar flow rate error (kmol/hr)
        self._mol_error = 0
        
        #: Temperature error (K)
        self._T_error = 0
        
        #: Specification error
        self._spec_error = 0
        
        #: Number of iterations
        self._iter = 0
         
        #: set[Stream] All streams within the system
        self.streams = streams = set()

        #: set[System] All subsystems in the system
        self.subsystems = subsystems = set()
        
        #: list[Unit] Network of only unit operations
        self._unitnetwork = units = []
        
        #: tuple[Unit, function and/or System] A network that is run element by element until the recycle converges.
        self.network = tuple(network)
        inst = isinstance
        for i in network:
            if i in units: continue
            if inst(i, Unit): 
                units.append(i)
                streams.update(i._ins)
                streams.update(i._outs)
            elif inst(i, System):
                units.extend(i._unitnetwork)
                subsystems.add(i)
                streams.update(i.streams)
        streams.discard(MissingStream) 
        
        # link all unit operations with linked streams
        for u in units:
            if inst(u, bst.Static): u.link_streams()
        
        #: set[Unit] All units within the system
        self.units = units = set(units)
        
        #: set[Unit] All units in the network that have costs
        self._network_costunits = costunits = {i for i in units if i._cost}
        
        #: set[Unit] All units that have costs.
        self._costunits = costunits = costunits.copy()
        
        #: tuple[Unit, function, and/or System] Offsite facilities that are simulated only after completing the network simulation.
        self.facilities = tuple(facilities)
        for i in facilities:
            if inst(i, Unit):
                units.add(i)
                streams.update(i._ins + i._outs)
                if i._cost: costunits.add(i)
                if inst(i, Facility): i._system = self
            elif inst(i, System):
                units.update(i.units)
                streams.update(i.streams)
                subsystems.add(i)
                costunits.update(i._costunits)
        
        #: set[Stream] All feed streams in the system.
        self.feeds = set(filter(_isfeed, streams))
        
        #: set[Stream] All product streams in the system.
        self.products = set(filter(_isproduct, streams)) 
        
        #: [TEA] System object for Techno-Economic Analysis.
        self._TEA = None
        
        if recycle is None:
            self._converge = self._run
        else:
            assert isinstance(recycle, Stream), (
             "recycle must be a Stream instance or None, not "
            f"{type(recycle).__name__}")
        self._recycle = recycle
        if ID: setattr(find.system, ID, self)
        else: self._ID = ""
    
    save_report = save_report
    
    @property
    def ID(self):
        """Identification."""
        return self._ID
    
    @property
    def TEA(self):
        """[TEA] System object for Techno-Economic Analysis."""
        return self._TEA
    
    @property
    def recycle(self):
        """A tear stream for the recycle loop"""
        return self._recycle

    @property
    def converge_method(self):
        """Iterative convergence method ('wegstein', 'aitken', or 'fixed point')."""
        return self._converge.__name__[1:]

    @converge_method.setter
    def converge_method(self, method):
        if self.recycle is None:
            raise ValueError("cannot set converge method when no recyle is specified")
        method = method.lower().replace('-', '').replace(' ', '')
        if 'wegstein' == method:
            self._converge = self._wegstein
        elif 'fixedpoint' == method:
            self._converge = self._fixed_point
        elif 'aitken' == method:
            self._converge = self._aitken
        else:
            raise ValueError(f"only 'wegstein', 'aitken', and 'fixed point' methods are valid, not '{method}'")

    
    def _downstream_network(self, unit):
        """Return a list composed of the `unit` and everything downstream."""
        if unit not in self.units: return []
        elif self._recycle: return self.network
        unit_found = False
        downstream_units = unit._downstream_units
        network = []
        inst = isinstance
        for i in self.network:
            if unit_found:
                if inst(i, System):
                    for u in i.units:
                        if u in downstream_units:
                            network.append(i)
                            break
                elif i in downstream_units:
                    network.append(i)
                elif (not inst(i, Unit)
                      or i.line == 'Balance'):
                    network.append(i)
            else:
                if unit is i:
                    unit_found = True
                    network.append(unit)
                elif inst(i, System) and unit in i.units:
                        unit_found = True   
                        network.append(i)
        return network

    def _downstream_system(self, unit):
        """Return a system with a network composed of the `unit` and everything downstream (facilities included)."""
        if unit is self.network[0]: return self
        system = self._cached_downstream_systems.get((self, unit))
        if system: return system
        network = self._downstream_network(unit)
        if network:
            downstream_facilities = self.facilities            
        else:
            unit_found = False
            inst = isinstance
            for pos, i in enumerate(self.facilities):
                if unit is i or (inst(i, System) and unit in i.units):
                    downstream_facilities = self.facilities[pos:]
                    unit_found = True
                    break
            assert unit_found, f'{unit} not found in system'
        system = System(None, network,
                        facilities=downstream_facilities)
        system._ID = f'{type(unit).__name__}-{unit} and downstream'
        self._cached_downstream_systems[unit] = system
        return system
    
    def _minimal_diagram(self, file, format, **graph_attrs):
        """Minimally display the network as a box."""
        outs = []
        ins = []
        for s in self.streams:
            source = s._source
            sink = s._sink
            if source in self.units and sink not in self.units:
                outs.append(s)
            elif sink in self.units and source not in self.units:
                ins.append(s)
        product = Stream(None)
        product._ID = ''
        feed = Stream(None)
        feed._ID = ''
        _streamUnit('\n'.join([i.ID for i in ins]),
                    None, feed)
        _streamUnit('\n'.join([i.ID for i in outs]),
                    product, None)
        unit = _systemUnit(self.ID, feed, product)
        unit.diagram(1, file=file, format=format, **graph_attrs)

    def _surface_diagram(self, file, format, **graph_attrs):
        """Display only surface elements listed in the network."""
        # Get surface items to make nodes and edges
        units = set()  
        refresh_units = set()
        for i in self.network:
            if isinstance(i, Unit):
                units.add(i)
            elif isinstance(i, System):
                outs = []
                ins = []
                feeds = []
                products = []
                for s in i.streams:
                    source = s._source
                    sink = s._sink
                    if source in i.units and sink not in i.units:
                        if sink: outs.append(s)
                        else: products.append(s)
                        refresh_units.add(source)
                    elif sink in i.units and source not in i.units:
                        if source: ins.append(s)
                        else: feeds.append(s)
                        refresh_units.add(sink)
                
                if len(feeds) > 1:
                    feed = Stream(None)
                    feed._ID = ''
                    units.add(_streamUnit('\n'.join([i.ID for i in feeds]), None, feed))
                    ins.append(feed)
                else: ins += feeds
                
                if len(products) > 1:
                    product = Stream(None)
                    product._ID = ''
                    units.add(_streamUnit('\n'.join([i.ID for i in products]),
                                          product, None))
                    outs.append(product)
                else: outs += products
                
                subsystem_unit = _systemUnit(i.ID, ins, outs)
                units.add(subsystem_unit)
                
        System(None, units)._thorough_diagram(file, format, **graph_attrs)
        # Reconnect how it was
        for u in refresh_units:
            u._ins[:] = u._ins
            u._outs[:] = u._outs
      
    def _thorough_diagram(self, file, format, **graph_attrs):
        """Thoroughly display every unit within the network."""
        # Create a digraph and set direction left to right
        f = make_digraph(self.units, self.streams, **graph_attrs)
        save_digraph(f, file, format)
        
    def diagram(self, kind='surface', file=None, format='png', **graph_attrs):
        """Display a `Graphviz <https://pypi.org/project/graphviz/>`__ diagram of the system.
        
        Parameters
        ----------
        kind='surface' : {'thorough', 'surface', 'minimal'}:
            * **'thorough':** Display every unit within the network.
            * **'surface':** Display only elements listed in the network.
            * **'minimal':** Display network as a box.
        file=None : str, display in console by default
            File name to save diagram.
        format='png' : str
            File format (e.g. "png", "svg").
        
        """
        if kind == 'thorough':
            return self._thorough_diagram(file, format, **graph_attrs)
        elif kind == 'surface':
            return self._surface_diagram(file, format, **graph_attrs)
        elif kind == 'minimal':
            return self._minimal_diagram(file, format, **graph_attrs)
        else:
            raise ValueError(f"kind must be either 'thorough', 'surface', or 'minimal'")
            
    # Methods for running one iteration of a loop
    def _iter_run(self, mol):
        """Run the system at specified recycle molar flow rate.
        
        Parameters
        ----------
        mol : numpy.ndarray
              Recycle molar flow rates.
            
        Returns
        -------
        rmol : numpy.ndarray
               New recycle molar flow rates.
        unconverged : bool
                      True if recycle has not converged.
            
        """
        recycle = self.recycle
        rmol = recycle.mol
        rmol[:] = mol
        T = recycle.T
        self._run()
        self._mol_error = abs(mol - recycle.mol).sum()
        self._T_error = abs(T - recycle.T)
        self._iter += 1
        if self._mol_error < self.molar_tolerance and self._T_error < self.T_tolerance:
            return rmol.copy(), False
        elif self._iter > self.maxiter:
            raise SolverError(f'{repr(self)} could not converge' + self._error_info())
        else:
            return rmol.copy(), True
        
    def _run(self):
        """Rigorous run each element of the system."""
        inst = isinstance
        _try = _try_method
        for a in self.network:
            if inst(a, Unit): _try(a._run)
            elif inst(a, System): a._converge()
            else: a() # Assume it is a function
    
    # Methods for convering the recycle stream
    def _fixed_point(self):
        """Converge system recycle using inner and outer loops with fixed-point iteration."""
        r = self.recycle
        rmol = r._mol
        while True:
            mol = rmol.copy()
            T = r.T
            self._run()
            self._mol_error = abs(mol - rmol).sum()
            self._T_error = abs(T - r.T)
            self._iter += 1
            if (self._mol_error < self.molar_tolerance
                and self._T_error < self.T_tolerance): break
            if self._iter > self.maxiter:
                raise SolverError(f'{repr(self)} could not converge' + self._error_info())
            
    def _wegstein(self):
        """Converge the system recycle iteratively using wegstein's method."""
        conditional_wegstein(self._iter_run, self.recycle._mol.copy())
    
    def _aitken(self):
        """Converge the system recycle iteratively using Aitken's method."""
        conditional_aitken(self._iter_run, self.recycle._mol.copy())
    
    # Default converge method
    _converge = _aitken

    def set_spec(self, getter, setter, solver=newton, **kwargs):
        """Wrap a solver around the converge method.

        Parameters
        ----------
        getter : function
            Returns objective value.
        setter : function
            Changes independent variable.
        solver=scipy.optimize.newton : function
            Solves objective function.
        **kwargs
            Key word arguments passed to solver.

        """
        converge = self._converge

        def error(val):
            setter(val)
            converge()
            self._spec_error = e = getter()
            return e

        def converge_spec():
            x = solver(error, **kwargs)
            if solver is newton: kwargs['x0'] = x

        self._converge = converge_spec

    def _reset_iter(self):
        self._iter = 0
        for system in self.subsystems: system._reset_iter()
    
    def reset_names(self, unit_format=None, stream_format=None):
        """Reset names of all streams and units according to the network order."""
        Unit._default_ID = unit_format if unit_format else ['U', 0]
        Stream._default_ID = stream_format if stream_format else ['d', 0]
        streams = set()
        units = set()
        for i in self._unitnetwork:
            if i in units: continue
            try: i.ID = ''
            except: continue
            for s in (i._ins + i._outs):
                if (s and s._sink and s._source
                    and s not in streams):
                    s.ID = ''
                    streams.add(s)
            units.add(i)
    
    def reset_flows(self):
        """Reset all process streams to zero flow."""
        self._iter = 0
        self._spec_error = 0
        self._mol_error = 0
        self._T_error = 0
        
        feeds = self.feeds
        for stream in self.streams:
            if stream._link: continue
            if stream not in feeds: stream.empty()

    def simulate(self):
        """Converge the network and simulate all units."""
        self._reset_iter()
        self._converge()
        _try = _try_method
        for u in self._network_costunits: _try(u._summary)
        inst = isinstance
        for i in self.facilities:
            if inst(i, (Unit, System)): i.simulate()
            else: i() # Assume it is a function
        
    # Debugging
    def _debug_on(self):
        """Turn on debug mode."""
        self._run = _notify_run_wrapper(self, self._run)
        self.network = network = list(self.network)
        for i, item in enumerate(network):
            if isinstance(item, Unit):
                item._run = _method_debug(item, item._run)
            elif isinstance(item, System):
                item._converge = _method_debug(item, item._converge)
            elif callable(item):
                network[i] = _method_debug(item, item)

    def _debug_off(self):
        """Turn off debug mode."""
        self._run = self._run._original
        network = self.network
        for i, item in enumerate(network):
            if isinstance(item, Unit):
                item._run = item._run._original
            elif isinstance(item, System):
                item._converge = item._converge._original
            elif callable(item):
                network[i] = item._original
        self.network = tuple(network)
    
    def debug(self):
        """Converge in debug mode. Just try it!"""
        self._debug_on()
        try: self._converge()
        finally: self._debug_off()
        end = self._error_info()
        if end:
            print(f'\nFinished debugging{end}')
        else:
            print(f'\n        Finished debugging')

    # Representation
    def __str__(self):
        if self.ID: return self.ID
        else: return type(self).__name__ 
    
    def __repr__(self):
        if self.ID: return f'<{type(self).__name__}: {self.ID}>'
        else: return f'<{type(self).__name__}>'

    def show(self):
        """Prints information on unit."""
        print(self._info())
    
    def _ipython_display_(self):
        try: self.diagram('minimal')
        except: pass
        self.show()

    def _error_info(self):
        """Return information on convergence."""
        recycle = self.recycle
        error_info = ''
        if self._spec_error:
            error_info += f'\n specification error: {self._spec_error:.3g}'
        if recycle:
            error_info +=(f"\n convergence error: Flow rate   {self._mol_error:.2e} kmol/hr"
                          f"\n                    Temperature {self._T_error:.2e} K")
        if self._spec_error or recycle:
            error_info += f"\n iterations: {self._iter}"
        return error_info

    def _info(self):
        """Return string with all specifications."""
        if self.recycle is None:
            recycle = ''
        else:
            recycle = f"\n recycle: {self.recycle}"
        error = self._error_info()
        network = strtuple(self.network)
        i = 1; last_i = 0
        while True:
            i += 2
            i = network.find(', ', i)
            i_next = network.find(', ', i+2)
            if (i_next-last_i) > 35:
                network = (network[:i] + '%' + network[i:])
                last_i = i
            elif i == -1: break
        network = network.replace('%, ', ',\n'+' '*11)
        
        if self.facilities:
            facilities = strtuple(self.facilities)
            i = 1; last_i = 0
            while True:
                i += 2
                i = facilities.find(', ', i)
                if (i - last_i) > 35:
                    facilities = (facilities[:i] + '%' + facilities[i:])
                    last_i = i
                elif i == -1: break
            facilities = facilities.replace('%, ', ',\n'+' '*14)
            facilities = f"\n facilities: {facilities}" 
        else:
            facilities = ''
        
        return (f"System: {self.ID}"
                + recycle
                + f"\n network: {network}"
                + facilities
                + error)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 18 16:17:00 2017

The biodiesel production section for the baseline lipid cane biorefinery is defined here as System objects. The systems include all streams and units starting from transesterification to purification of biodiesel and glycerol products.

@author: yoel
"""
from biosteam import System, Stream, MixedStream
from biosteam.units import Transesterification, \
    SplitFlash, MixTank, Distillation, RatioCentrifuge_LLE, \
    InvSplitter, Pump, StorageTank, \
    HXutility, HXprocess, Flash, \
    SplitCentrifuge_LLE, MassBalance
from biosteam.biorefineries.lipidcane.species import biodiesel_species
from biosteam.biorefineries.lipidcane.process_settings import price
from numpy import array


__all__ = ('area_400',)

# %% Stream Settings

# Set species library
Stream.species = biodiesel_species
# Stream.default_ID_number = 400


# %% Fresh Streams

# Fresh degummed oil
oil = Stream('oil', Lipid=8853.49, Water=1.00e-02,
             units='kg/hr', T=347.15)

# Fresh methanol
methanol = Stream('methanol', Methanol=1,
                  price=price['Methanol'])

# Catalyst
catalyst = Stream('catalyst', NaOCH3=0.25,
                  Methanol=0.75, units='kg/hr',
                  price=price['NaOCH3'])
                  # price=0.25*price['NaOCH3'] + 0.75*Methanol.price)

# Water to remove glycerol
biodiesel_wash_water = Stream('biodiesel_wash_water', Water=13.6, T=273.15+60, 
               price=price['Water'])

# Acid to neutralize catalyst after second centrifuge
HCl1 = Stream('HCl1', HCl=0.21, Water=0.79,
              price=price['HCl'])
              # price=0.21*price['HCl'] + 0.79*Water.price) # 35% HCl by mass

# Acid to remove soaps after first centrifuge
HCl2 = Stream('HCl2', HCl=0.21, Water=0.79,
              price=HCl1.price)

# Base to neutralize acid before distillation
NaOH = Stream('NaOH', NaOH=1, price=price['NaOH'])

# Products
crude_glycerol = Stream('crude_glycerol',
                        price=price['Crude glycerol'])
biodiesel = Stream('biodiesel',
                   price=price['Biodiesel'])

# Waste
waste = Stream('waste', price=price['Waste'])

# %% Units

"""Biodiesel Transesterification Section"""

# Aparently reactors are adiabatic, meoh coming in at 40C, lipid at 60C

# From USDA Biodiesel model
x_cat = 1.05e-05 # Catalyst molar fraction in methanol feed

# Mix Recycle Methanol and Fresh Methanol
T401 = StorageTank('T401')
P401 = Pump('P401')

# Storage Tank for Catalyst
T402 = StorageTank('T402')
P402 = Pump('P402')

# Tank for oil
T403 = StorageTank('T403')
T403.tau = 4
P403 = Pump('P403')

# Mix Methanol and Catalyst stream
T404 = MixTank('T404')
P404 = Pump('P404')

# Mass Balance for Methanol, Recycle Methanol, and Catalyst stream
B401 = MassBalance('B401', species=('Methanol', 'NaOCH3'), streams=(0, 1))

# Split Methanol/Catalyst to reactors
S401 = InvSplitter('S401')

# First Reactor
R401 = Transesterification('R401', efficiency=0.90, methanol2lipid=6, T=333.15,
                         catalyst_molfrac=x_cat)

# Centrifuge to remove glycerol
C401 = SplitCentrifuge_LLE('C401',
                         split=(0.99,  # Lipid
                                0.40,  # Methanol
                                0.06,  # Glycerol
                                0.999, # Biodiesel
                                0.40,  # Water
                                0,     # NaOH
                                0,     # HCl
                                0.40)) # NaOCH3

P405 = Pump('P405')

# Second Reactor
R402 = Transesterification('R402', efficiency=0.90, methanol2lipid=6, T=333.15,
                         catalyst_molfrac=x_cat) 

# Centrifuge to remove glycerol
C402 = SplitCentrifuge_LLE('C402',
                         split=(0.90,  # Lipid
                                0.10,  # Methanol
                                0.05,  # Glycerol
                                0.999, # Biodiesel
                                0.10,  # Water
                                0,     # NaOH
                                0,     # HCl
                                0.10)) # NaOCH3

# Acids and bases per catalyst by mol
catalyst_index = oil.index('NaOCH3')
k1 = 0.323/1.5; k2 = 1.060/1.5; k3 = 0.04505/1.5;
catalyst_mol = T404.outs[0].mol[7]
def adjust_acid_and_base():
    global catalyst_mol
    # Adjust according to USDA biodiesel model
    new = T404.outs[0].mol[7]
    if new != catalyst_mol:
        catalyst_mol = new
        NaOH._mol[5] = k1 * new
        HCl1._mol[6] = k2 * new
        HCl2._mol[6] = k3 * new

"""Biodiesel Purification Section"""

# Water wash
T405 = MixTank('T405')
T405.tau = 0.25
P406 = Pump('P406')

# Centrifuge out water
C403 = RatioCentrifuge_LLE('C403',
                         Kspecies=('Methanol', 'Glycerol'),
                         Ks=array((0.382, 0.183)),
                         top_solvents=('Biodiesel',),
                         top_split=(0.999,),
                         bot_solvents=('Water', 'Lipid', 'NaOH', 'HCl'),
                         bot_split=(0.999, 1, 1, 1))

# Vacuum dry biodiesel
# Consider doing this by split, and keeping this unit constant
# 290 Pa, 324 K according to USDA Biodiesel Model
F401 = SplitFlash('F401',
                order=('Water', 'Methanol', 'Biodiesel'),
                split=(0.9999, 0.9999, 0.00001),
                P=2026.5, T=331.5, Q=0)
F401.line = 'Vacuum dryer'
F401.material = 'Stainless steel 316'
P407 = Pump('P407')

"""Glycerol Purification Section"""

# Condense vacuumed methanol to recycle
H401 = HXutility('H401', V=0, T=295)
P408 = Pump('P408')

# Mix recycled streams and HCl
T406 = MixTank('T406')
P409 = Pump('P409')

# Centrifuge out waste fat
# assume all the lipid, free_lipid and biodiesel is washed out
C404 = SplitCentrifuge_LLE('C404', outs=('', waste),
                         order=('Methanol', 'Glycerol', 'Water'),
                         split=(0.999, 0.999, 0.999))

# Add and mix NaOH
T407 = MixTank('T407')
P410 = Pump('P410')

# Methanol/Water distillation column
D401 = Distillation('D401',
                  LHK=('Methanol', 'Water'), P=101325,
                  y_top=0.99999, x_bot=0.0001, k=2.5)
D401.is_divided = True
D401.tray_material = 'Stainless steel 316'
D401.vessel_material = 'Stainless steel 316'

# Condense recycle methanol
H402 = HXutility('H402', V=0, T=315)
P411 = Pump('P411')

# Glycerol/Water flash (not a distillation column)
w = 0.20/Stream.species.Water.MW
g = 0.80/Stream.species.Glycerol.MW
x = w/(w+g)
F402 = Flash('F402',
           species_IDs=('Water', 'Glycerol'),
           P=101325,
           x=(x, 1-x),
           HNK=('Biodiesel',),
           LNK=('Methanol',))
F402.HasGlycolGroups = True
F402.vessel_material = 'Stainless steel 316'

# Condense water to recycle
H403 = HXprocess('H403', fluid_type='ll', outs=(MixedStream(), ''),
                     species_IDs=('Methanol', 'Water'), 
                     HNK=('Glycerol', 'Biodiesel', 'NaOH', 'HCl'))
    
# Use heat from glycerol
H404 = HXprocess('H404', fluid_type='ls',
                     species_IDs=('Methanol', 'Water'), 
                     HNK=('Glycerol', 'Biodiesel', 'NaOH', 'HCl'))
run_bot = H404._run

# Storage tank for glycerol
T408 = StorageTank('T408', outs=crude_glycerol)
H404-1-T408

# Storage tank for biodiesel
T409 = StorageTank('T409', outs=biodiesel)
F401-1-P407-0-T409

# %% Set up systems

# Biodiesel Transesterification Section
oil-T403-P403
(P403-0, S401-0)-R401-0-C401
(C401-0, S401-1)-R402-0-C402
transesterification_network = (T403, P403, R401, C401, P405, R402, C402,
                               adjust_acid_and_base)

"""
Specs for product https://www.afdc.energy.gov/fuels/biodiesel_specifications.html
minimum spec requirements:
 0.050 wt % water (lower to 0.045)
 0.2 wt % meoh (lower to 0.15)
 0.02 wt % glycerol (lower to 0.015)
 0.4 wt % free lipids (lower to 0.35)
"""

# Find Water Flow
(C402-0, H403-0, biodiesel_wash_water, HCl1)-T405-P406-C403
def adjust_water_flow():
    H403.outs[0].liquid_mol[4] = 800*C402.outs[0]._mol[2]

# Glycerol recycle and purification section
C403-0-F401
F401-0-H401-P408
C401-1-P405
(P405-0, C402-1, C403-1, P408-0, HCl2)-T406-P409-C404
(C404-0, NaOH)-T407-P410
(H403-1, F402-1)-H404
H404-0-D401-1-F402
(F402-0, P410-0)-H403
glycerol_recycle_sys = System('glycerol_recycle_sys',
                              network=(adjust_water_flow, 
                                       T405, P406, C403, F401, H401,
                                       P408, P407, T406, P409, C404,
                                       T407, P410, H403, H404, D401,
                                       F402, H403, H404),
                              recycle=F402-0)                           
# Find Fresh Methanol Flow
D401-0-H402-P411    # Recycle methanol
methanol-T401-P401  # Mix fresh and recycled methanol
catalyst-T402-P402  # Fresh catalyst
(P411-0, P401-0, P402-0)-T404-P404-S401  # Mix Catalyst with Methanol
meoh_network = (H402, P411, T401, P401, T402, P402, T404, P404, S401)

# Set connections for mass balance proxy
(catalyst, methanol, D401-0)-B401
B401**(1**R401, 1**R402)

# Complete System
area_400 = System('area_400',
                  network=transesterification_network
                          + (glycerol_recycle_sys, B401)
                          + meoh_network
                          + (T408, T409))

# Initial conditions
index = oil.indices(['Methanol', 'Glycerol', 'Water'])
F402.outs[0].mol[index] = (3.483e-02, 8.284e-03, 2.094e+01)
F402.outs[0].T = 374.986
H403.outs[0].T = 374.986
H403.outs[0].liquid_mol[index] = (3.483e-02, 8.284e-03, 2.094e+01)

